"""Functions related to creating, deleting and keeping track of physical Splitgraph objects."""
import datetime
import itertools
import logging
import math
from collections import defaultdict
from contextlib import contextmanager
from datetime import datetime as dt
from typing import (
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from psycopg2.sql import SQL, Identifier

from splitgraph.config import CONFIG, SPLITGRAPH_META_SCHEMA, get_singleton
from splitgraph.core.fragment_manager import FragmentManager
from splitgraph.core.types import Quals
from splitgraph.engine import ResultShape, switch_engine
from splitgraph.exceptions import (
    IncompleteObjectDownloadError,
    IncompleteObjectUploadError,
    ObjectCacheError,
    SplitGraphError,
)
from splitgraph.hooks.external_objects import get_external_object_handler

from .common import META_TABLES, CallbackList, Tracer
from .output import pluralise, pretty_size, truncate_list
from .sql import insert, select

if TYPE_CHECKING:
    from splitgraph.core.table import Table
    from splitgraph.engine.postgres.engine import PostgresEngine, PsycopgEngine


class ObjectManager(FragmentManager):
    """Brings the multiple manager classes together and manages the object cache (downloading and uploading
    objects as required in order to fulfill certain queries)"""

    def __init__(
        self, object_engine: "PostgresEngine", metadata_engine: Optional["PostgresEngine"] = None
    ) -> None:
        """
        :param object_engine: An ObjectEngine that will be used as a backing store for the
            objects.
        :param metadata_engine: An SQLEngine that will be used to store/query metadata for Splitgraph
            images and objects. By default, `object_engine` is used.
        """
        super().__init__(object_engine, metadata_engine)

        # Cache size in bytes
        self.cache_size = int(get_singleton(CONFIG, "SG_OBJECT_CACHE_SIZE")) * 1024 * 1024

        # 0 to infinity; higher means objects with smaller sizes are more likely to
        # get evicted than objects that haven't been used for a while.
        # Currently calculated so that an object that hasn't been accessed for 5 minutes has the same
        # removal priority as an object twice its size that's just been accessed.
        self.eviction_decay_constant = float(get_singleton(CONFIG, "SG_EVICTION_DECAY"))

        # Objects smaller than this size are assumed to have this size (to simulate the latency of
        # downloading them).
        self.eviction_floor = float(get_singleton(CONFIG, "SG_EVICTION_FLOOR")) * 1024 * 1024

        # Fraction of the cache size to free when eviction is run (the greater value of this amount and the
        # amount needed to download required objects is actually freed). Eviction is an expensive operation
        # (it pauses concurrent downloads) so increasing this makes eviction happen less often at the cost
        # of more possible cache misses.
        self.eviction_min_fraction = float(get_singleton(CONFIG, "SG_EVICTION_MIN_FRACTION"))

    def get_downloaded_objects(self, limit_to: Optional[List[str]] = None) -> List[str]:
        """
        Gets a list of objects currently in the Splitgraph cache (i.e. not only existing externally.)

        :param limit_to: If specified, only the objects in this list will be returned.
        :return: Set of object IDs.
        """
        objects: List[str]
        objects = self.object_engine.run_api_call("list_objects")
        if not limit_to:
            return objects
        return [o for o in objects if o in limit_to]

    def get_cache_occupancy(self) -> int:
        """
        :return: Space occupied by objects cached from external locations, in bytes.
        """
        return int(
            self.object_engine.run_sql(
                SQL("SELECT total_size FROM {}.object_cache_occupancy").format(
                    Identifier(SPLITGRAPH_META_SCHEMA)
                ),
                return_shape=ResultShape.ONE_ONE,
            )
        )

    def _recalculate_cache_occupancy(self) -> int:
        """A slower way of getting cache occupancy that actually goes through all objects in the cache status table
        and sums up their size."""
        return int(
            self.object_engine.run_sql(
                SQL(
                    "SELECT COALESCE(sum(splitgraph_api.get_object_size("
                    "quote_ident(t.table_name))), 0)"
                    " FROM information_schema.tables t JOIN {0}.object_cache_status oc"
                    " ON t.table_name = oc.object_id"
                    " WHERE oc.ready = 't' AND t.table_schema = %s AND t.table_type = 'FOREIGN'"
                ).format(Identifier(SPLITGRAPH_META_SCHEMA)),
                (SPLITGRAPH_META_SCHEMA,),
                return_shape=ResultShape.ONE_ONE,
            )
        )

    def get_total_object_size(self):
        """
        :return: Space occupied by all objects on the engine, in bytes.
        """
        return int(
            self.object_engine.run_sql(
                SQL(  # nosec
                    "SELECT COALESCE(sum(splitgraph_api.get_object_size("
                    "quote_ident(t.table_name))), 0)"
                    " FROM information_schema.tables t"
                    " WHERE t.table_schema = %s AND t.table_name NOT IN ("
                    + ",".join(itertools.repeat("%s", len(META_TABLES)))
                    + ")"
                ).format(Identifier(SPLITGRAPH_META_SCHEMA)),
                [SPLITGRAPH_META_SCHEMA] + META_TABLES,
                return_shape=ResultShape.ONE_ONE,
            )
        )

    @contextmanager
    def ensure_objects(
        self,
        table: Optional["Table"],
        objects: Optional[List[str]] = None,
        quals: Optional[Quals] = None,
        defer_release: bool = False,
        tracer: Optional[Tracer] = None,
        upstream_manager: Optional["ObjectManager"] = None,
    ) -> Iterator[Union[List[str], Tuple[List[str], CallbackList]]]:
        """
        Resolves the objects needed to materialize a given table and makes sure they are in the local
        splitgraph_meta schema.

        Whilst inside this manager, the objects are guaranteed to exist. On exit from it, the objects are marked as
        unneeded and can be garbage collected.

        :param table: Table to materialize
        :param objects: List of objects to download: one of table or objects must be specified.
        :param quals: Optional list of qualifiers to be passed to the fragment engine. Fragments that definitely do
            not match these qualifiers will be dropped. See the docstring for `filter_fragments` for the format.
        :param defer_release: If True, won't release the objects on exit.
        :return: If defer_release is True: List of table fragments and a callback that the caller must call
            when the objects are no longer needed. If defer_release is False: just the list of table fragments.
        """

        # Main cache management issue here is concurrency: since we have multiple processes per connection on the
        # server side, if we're accessing this from the FDW, multiple ObjectManagers can fiddle around in the cache
        # status table, triggering weird concurrency cases like:
        #   * We need to download some objects -- another manager wants the same objects. How do we make sure we don't
        #     download them twice? Do we wait on a row-level lock? What happens if another manager crashes?
        #   * We have decided which objects we need -- how do we make sure we don't get evicted by another manager
        #     between us making that decision and increasing the refcount?
        #   * What happens if we crash when we're downloading these objects?

        self.object_engine.run_sql("SET LOCAL synchronous_commit TO OFF")
        tracer = tracer or Tracer()

        if objects is not None:
            required_objects = objects
        else:
            if table is None:
                raise ValueError("One of table or objects must be specified!")
            logging.debug(
                "Resolving objects for table %s:%s:%s",
                table.repository,
                table.image.image_hash,
                table.table_name,
            )

            # Filter to see if we can discard any objects with the quals
            required_objects = self.filter_fragments(table.objects, table, quals)
            tracer.log("filter_objects")

        # Increase the refcount on all of the objects we're giving back to the caller so that others don't GC them.
        logging.debug("Claiming %s", pluralise("object", len(required_objects)))

        self._claim_objects(required_objects)
        tracer.log("claim_objects")

        try:
            to_fetch = self._prepare_fetch_list(required_objects)
        except SplitGraphError:
            self.object_engine.rollback()
            raise
        tracer.log("prepare_fetch_list")

        # Perform the actual download. If the table has no upstream but still has external locations, we download
        # just the external objects.
        if to_fetch:
            object_locations = self.get_external_object_locations(to_fetch)

            # If all objects are externally hosted, there's no need to try and get the table's
            # upstream (there's a corner case where the metadata engine is different from the object
            # engine and the repo actually has no upstream)
            if (
                self.metadata_engine == self.object_engine
                and table is not None
                and upstream_manager is None
            ):
                upstream_manager = (
                    table.repository.upstream.objects if table.repository.upstream else None
                )

            partial_failure: Optional[BaseException] = None
            try:
                successful = self.download_objects(
                    upstream_manager, objects_to_fetch=to_fetch, object_locations=object_locations
                )
                difference = []
            except IncompleteObjectDownloadError as e:
                successful = e.successful_objects
                difference = list(set(to_fetch).difference(successful))
                if difference:
                    if e.reason:
                        partial_failure = e.reason
                    else:
                        partial_failure = self._generate_download_error(table, difference)
            except Exception as e:
                successful = self.get_downloaded_objects(to_fetch)
                difference = list(set(to_fetch).difference(successful))
                partial_failure = self._generate_download_error(table, difference, cause=e)

            # No matter what, claim the space required by the newly downloaded objects.
            if successful:
                self._increase_cache_occupancy(successful)

            if partial_failure:
                # Instead of deleting all objects in this batch, discard the cache data
                # on the objects that failed, decrease the refcount on the objects that
                # succeeded and mark them as ready.
                self._delete_cache_entries(difference)
                self._set_ready_flags(successful, is_ready=True)
                self._release_objects(successful)
                self.object_engine.commit()

                raise partial_failure

            self._set_ready_flags(to_fetch, is_ready=True)
        tracer.log("fetch_objects")
        logging.debug("Object manager finished.")

        release_callback = self._make_release_callback(required_objects, table, tracer)
        try:
            # Release the lock and yield to the caller.
            self.object_engine.commit()
            self.metadata_engine.commit()

            if defer_release:
                yield required_objects, release_callback
            else:
                yield required_objects
        finally:
            if not defer_release:
                release_callback()

    def _generate_download_error(self, table, difference, cause=None):
        if table:
            error = "Not all objects required for %s:%s:%s have been fetched. Missing %s (%s)" % (
                table.repository.to_schema(),
                table.image.image_hash,
                table.table_name,
                pluralise("object", len(difference)),
                truncate_list(difference),
            )
        else:
            error = "Not all objects have been fetched. Missing %s (%s)" % (
                pluralise("object", len(difference)),
                truncate_list(difference),
            )
        logging.error(error)
        partial_failure = ObjectCacheError(error)
        if cause:
            partial_failure.__cause__ = cause
        return partial_failure

    def _make_release_callback(
        self, required_objects: List[str], table: Optional["Table"], tracer: Tracer
    ) -> CallbackList:
        called = False

        def _f(**kwargs):
            nonlocal called
            if called:
                return
            called = True
            # Decrease the refcounts on the objects. Optionally, evict them.
            # If it's the caller's responsibility to call this and it crashes,
            # we'll leak memory (hold on to objects in the cache that could have been
            # garbage collected).
            tracer.log("caller")
            self.object_engine.run_sql("SET LOCAL synchronous_commit TO off")
            self._release_objects(required_objects)
            tracer.log("release_objects")
            logging.debug("Releasing %s", pluralise("object", len(required_objects)))
            if table:
                logging.debug(
                    "Timing stats for %s/%s/%s/%s: \n%s",
                    table.repository.namespace,
                    table.repository.repository,
                    table.image.image_hash,
                    table.table_name,
                    tracer,
                )
            self.object_engine.commit()
            # Release the metadata tables as well
            self.metadata_engine.commit()

        return CallbackList([_f])

    def make_objects_external(
        self, objects: List[str], handler: str, handler_params: Dict[Any, Any]
    ) -> None:
        """
        Uploads local objects to an external location and marks them as being cached locally (thus making it possible
        to evict or swap them out).

        :param objects: Object IDs to upload. Will do nothing for objects that already exist externally.
        :param handler: Object handler
        :param handler_params: Extra handler parameters
        """
        # Get objects that haven't been uploaded
        uploaded_objects = [o[0] for o in self.get_external_object_locations(objects)]
        new_objects = [o for o in objects if o not in uploaded_objects]

        logging.debug(
            "%s of %d haven't been uploaded yet: %r",
            pluralise("object", len(new_objects)),
            len(objects),
            new_objects,
        )

        if not new_objects:
            return

        # Similar to claim_objects, make sure we don't deadlock with other uploaders
        # by keeping a consistent order.
        new_objects = sorted(new_objects)

        # Insert the objects into the cache status table (marking them as not ready)
        now = dt.utcnow()
        self.object_engine.run_sql_batch(
            insert("object_cache_status", ("object_id", "ready", "refcount", "last_used"))
            + SQL("ON CONFLICT (object_id) DO UPDATE SET ready = 'f'"),
            [(object_id, False, 1, now) for object_id in new_objects],
        )

        # Grab the objects that we're supposed to be uploading.
        claimed_objects = self.object_engine.run_sql(
            select("object_cache_status", "object_id", "ready = 'f' FOR UPDATE"),
            return_shape=ResultShape.MANY_ONE,
        )
        new_objects = [o for o in new_objects if o in claimed_objects]

        # Perform the actual upload
        external_handler = get_external_object_handler(handler, handler_params)

        partial_failure: Optional[IncompleteObjectUploadError] = None
        try:
            with switch_engine(self.object_engine):
                successful = {
                    o: u
                    for o, u in external_handler.upload_objects(new_objects, self.metadata_engine)
                }
        except IncompleteObjectUploadError as e:
            partial_failure = e
            successful = {o: u for o, u in zip(e.successful_objects, e.successful_object_urls)}

        locations = [(o, u, handler) for o, u in successful.items()]
        self.register_object_locations(locations)

        # Increase the cache occupancy since the objects can now be evicted.
        self._increase_cache_occupancy(list(successful.keys()))

        # Mark the objects as ready and decrease their refcounts.
        self._set_ready_flags(list(successful.keys()), True)
        self._release_objects(list(successful.keys()))

        self.object_engine.commit()
        self.metadata_engine.commit()

        if partial_failure:
            if partial_failure.reason:
                raise partial_failure.reason
            else:
                raise ObjectCacheError("Some objects failed to upload!")

        # Perform eviction in case we've reached the capacity of the cache
        excess = self.get_cache_occupancy() - self.cache_size
        if excess > 0:
            self.run_eviction(keep_objects=[], required_space=excess)

    def _prepare_fetch_list(self, required_objects: List[str]) -> List[str]:
        """
        Calculates the missing objects and ensures there's enough space in the cache
        to download them.

        :param required_objects: Iterable of object IDs that are required to be on the engine.
        :return: Set of objects to fetch
        """
        to_fetch: List[str] = self.object_engine.run_sql(
            select("object_cache_status", "object_id", "ready = 'f'"),
            return_shape=ResultShape.MANY_ONE,
        )
        if to_fetch:
            # If we need to download anything, take out an exclusive lock on the cache since we might
            # need to run eviction and don't want multiple managers trying to download the same things.
            # This used to be more granular (allow multiple managers downloading objects) but was resulting
            # in fun concurrency bugs and deadlocks that I don't have the willpower to investigate further
            # right now (e.g. two managers trying to download the same objects at the same time after one of them
            # runs cache eviction and releases some locks -- or two managers trying to free the same amount
            # of space in the cache for the same set of objects).

            # Since we already hold a row-level lock on some objects in the cache, we have to release it first and
            # lock the full table then -- but this means someone else might start downloading the objects that
            # we claimed. So, once we acquire the lock, we recalculate the fetch list again to see what
            # we're supposed to be fetching.
            self.object_engine.commit()
            self.object_engine.lock_table(SPLITGRAPH_META_SCHEMA, "object_cache_status")
            to_fetch = self.object_engine.run_sql(
                select("object_cache_status", "object_id", "ready = 'f'"),
                return_shape=ResultShape.MANY_ONE,
            )
            # If someone else downloaded all the objects we need, there's no point in holding the lock.
            # This is tricky to test with a single process.
            if not to_fetch:  # pragma: no cover
                self.object_engine.commit()
                return to_fetch
            required_space = sum(o.size for o in self.get_object_meta(list(to_fetch)).values())
            current_occupied = self.get_cache_occupancy()
            logging.info(
                "Need to download %s (%s), cache occupancy: %s/%s",
                pluralise("object", len(required_objects)),
                pretty_size(required_space),
                pretty_size(current_occupied),
                pretty_size(self.cache_size),
            )
            # If the total cache size isn't large enough, there's nothing we can do without cooperating with the
            # caller and seeing if they can use the objects one-by-one.
            if required_space > self.cache_size:
                raise ObjectCacheError(
                    "Not enough space in the cache to download the required objects!"
                )
            if required_space > self.cache_size - current_occupied:
                to_free = required_space + current_occupied - self.cache_size
                logging.info("Need to free %s", pretty_size(to_free))
                self.run_eviction(required_objects, to_free)
            self.object_engine.commit()
            # Finally, after we're done with eviction, relock the objects that we're supposed to be downloading.
            to_fetch = self.object_engine.run_sql(
                select("object_cache_status", "object_id", "ready = 'f' FOR UPDATE"),
                return_shape=ResultShape.MANY_ONE,
            )
        return to_fetch

    def _claim_objects(self, objects: List[str]) -> None:
        """Increases refcounts and bumps the last used timestamp to now for cached objects.
        For objects that aren't in the cache, checks that they don't already exist locally and then
        adds them to the cache status table, marking them with ready=False
        (which must be set to True by the end of the operation)."""
        if not objects:
            return
        now = dt.utcnow()
        # Objects that were created locally aren't supposed to be claimed here or have an entry in the cache.
        # So, we first try to update cache entries to bump their refcount, see which ones we updated,
        # subtract objects that we have locally and insert the remaining entries as new cache entries.

        claimed = self.object_engine.run_sql(
            SQL(  # nosec
                "UPDATE {}.object_cache_status SET refcount = refcount + 1, "
                "last_used = %s WHERE object_id IN ("
            ).format(Identifier(SPLITGRAPH_META_SCHEMA))
            + SQL(",".join(itertools.repeat("%s", len(objects))))
            + SQL(") RETURNING object_id"),
            [now] + objects,  # type: ignore
            return_shape=ResultShape.MANY_ONE,
        )
        claimed = claimed or []
        remaining = set(objects).difference(set(claimed))
        remaining = remaining.difference(set(self.get_downloaded_objects(limit_to=list(remaining))))

        # Since we send multiple queries, each claiming a single remaining object, we can deadlock here
        # with another object manager instance. Hence, we sort the list of objects so that we claim them
        # in a consistent order between all instances.
        remaining = sorted(remaining)

        # Remaining: objects that are new to the cache and that we'll need to download. However, between us
        # running the first query and now, somebody else might have started downloading them. Hence, when
        # we try to insert them, we'll be blocked until the other engine finishes its download and commits
        # the transaction -- then get an integrity error. So here, we do an update on conflict (again).
        self.object_engine.run_sql_batch(
            insert("object_cache_status", ("object_id", "ready", "refcount", "last_used"))
            + SQL(
                "ON CONFLICT (object_id) DO UPDATE SET refcount = EXCLUDED.refcount + 1, last_used = %s"
            ),
            [(object_id, False, 1, now, now) for object_id in remaining],
        )

    def _set_ready_flags(self, objects: List[str], is_ready: bool = True) -> None:
        if objects:
            self.object_engine.run_sql(
                SQL(  # nosec
                    "UPDATE {0}.object_cache_status SET ready = %s WHERE object_id IN ("
                    + ",".join(itertools.repeat("%s", len(objects)))
                    + ")"
                ).format(Identifier(SPLITGRAPH_META_SCHEMA)),
                [is_ready] + list(objects),  # type: ignore
            )

    def _release_objects(self, objects: List[str]) -> None:
        """Decreases objects' refcounts."""
        if objects:
            self.object_engine.run_sql(
                SQL(  # nosec
                    "UPDATE {}.{} SET refcount = refcount - 1 WHERE object_id IN ("
                    + ",".join(itertools.repeat("%s", len(objects)))
                    + ")"
                ).format(Identifier(SPLITGRAPH_META_SCHEMA), Identifier("object_cache_status")),
                objects,
            )

    def _increase_cache_occupancy(self, objects: List[str]) -> None:
        """Increase the cache occupancy by objects' total size."""
        if not objects:
            return
        total_size = sum(o.size for o in self.get_object_meta(objects).values())
        self.object_engine.run_sql(
            SQL("UPDATE {}.object_cache_occupancy SET total_size = total_size + %s").format(
                Identifier(SPLITGRAPH_META_SCHEMA)
            ),
            (total_size,),
        )

    def _decrease_cache_occupancy(self, size_freed: int) -> None:
        """Decrease the cache occupancy by a given size."""
        self.object_engine.run_sql(
            SQL("UPDATE {}.object_cache_occupancy SET total_size = total_size - %s").format(
                Identifier(SPLITGRAPH_META_SCHEMA)
            ),
            (size_freed,),
        )

    def run_eviction(self, keep_objects: List[str], required_space: Optional[int] = None) -> None:
        """
        Delete enough objects with zero reference count (only those, since we guarantee that whilst refcount is >0,
        the object stays alive) to free at least `required_space` in the cache.

        :param keep_objects: List of objects (besides those with nonzero refcount) that can't be deleted.
        :param required_space: Space, in bytes, to free. If the routine can't free at least this much space,
            it shall raise an exception. If None, removes all eligible objects.
        """

        logging.info("Performing eviction...")
        # Find deletion candidates: objects that we have locally, with refcount 0, that aren't in the whitelist.

        candidates = [
            o
            for o in self.object_engine.run_sql(
                select("object_cache_status", "object_id,last_used", "refcount=0"),
                return_shape=ResultShape.MANY_MANY,
            )
            if o[0] not in keep_objects
        ]

        object_meta = self.get_object_meta([o[0] for o in candidates]) if candidates else {}
        object_sizes = {o.object_id: o.size for o in object_meta.values()}

        # Also delete objects that don't have a metadata entry at all
        orphaned_objects = [o[0] for o in candidates if o[0] not in object_sizes]
        orphaned_object_sizes = {o: self.object_engine.get_object_size(o) for o in orphaned_objects}
        if orphaned_objects:
            logging.info(
                "Found %s, total size %s: %s",
                pluralise("orphaned object", len(orphaned_objects)),
                pretty_size(sum(orphaned_object_sizes.values())),
                orphaned_objects,
            )

        if required_space is None:
            # Just delete everything with refcount 0.
            to_delete = [o[0] for o in candidates]
            freed_space = sum(object_sizes.values()) + sum(orphaned_object_sizes.values())
            logging.info(
                "Will delete %s, total size %s",
                pluralise("object", len(to_delete)),
                pretty_size(freed_space),
            )
        else:
            if required_space > sum(object_sizes.values()) + sum(orphaned_object_sizes.values()):
                raise ObjectCacheError("Not enough space will be reclaimed after eviction!")

            to_delete, freed_space = self._prepare_eviction_candidates(
                candidates, object_sizes, orphaned_object_sizes, orphaned_objects, required_space
            )

        if to_delete:
            # NB delete_objects commits as well, releasing the lock. Make sure to do all bookkeeping first so that
            # other object managers in this function think that the objects have been deleted and don't try to delete
            # them again.
            self._delete_cache_entries(to_delete)
            self._decrease_cache_occupancy(freed_space)
            self.delete_objects(to_delete)
            logging.info(
                "Eviction done. Cache occupancy: %s", pretty_size(self.get_cache_occupancy())
            )

    def _eviction_score(
        self, now: datetime.datetime, object_size: int, last_used: datetime.datetime
    ) -> float:
        # We want to evict objects in order to minimize
        # P(object is requested again) * (cost of redownloading the object).
        # To approximate the probability, we use an exponential decay function (1 if last_used = now, dropping down
        # to 0 as time since the object's last usage time passes).
        # To approximate the cost, we use the object's size, floored to a constant (so if the object has
        # size <= floor, we'd use the floor value -- this is to simulate the latency of re-fetching the object,
        # as opposed to the bandwidth)
        time_since_used = (now - last_used).total_seconds()
        time_factor = math.exp(-self.eviction_decay_constant * time_since_used)
        size_factor = object_size if object_size > self.eviction_floor else self.eviction_floor
        return time_factor * size_factor

    def _prepare_eviction_candidates(
        self, candidates, object_sizes, orphaned_object_sizes, orphaned_objects, required_space
    ):
        now = dt.utcnow()

        # Since we can free the minimum required amount of space, see if we can free even more as
        # per our settings (if we can't, we'll just delete as much as we can instead of failing).
        required_space = max(required_space, int(self.eviction_min_fraction * self.cache_size))

        # Delete all orphaned objects first
        to_delete = orphaned_objects
        last_useds = [o[1] for o in candidates if o[0] in orphaned_objects]
        freed_space = sum(orphaned_object_sizes.values())

        # Sort candidates by deletion priority (lowest is smallest expected retrieval cost -- more likely to delete)
        candidates = sorted(
            [o for o in candidates if o[0] not in orphaned_objects],
            key=lambda o: self._eviction_score(now, object_sizes[o[0]], o[1]),
        )

        # Keep adding deletion candidates until we've freed enough space.
        for object_id, last_used in candidates:
            if freed_space >= required_space:
                break
            last_useds.append(last_used)
            to_delete.append(object_id)
            freed_space += object_sizes[object_id]
        logging.info(
            "Will delete %s last used between %s and %s, total size %s: %s",
            pluralise("object", len(to_delete)),
            min(last_useds).isoformat(),
            max(last_useds).isoformat(),
            pretty_size(freed_space),
            to_delete,
        )
        return to_delete, freed_space

    def _delete_cache_entries(self, to_delete: List[str]) -> None:
        if not to_delete:
            return
        self.object_engine.run_sql(
            SQL(  # nosec
                "DELETE FROM {}.{} WHERE object_id IN ("
                + ",".join(itertools.repeat("%s", len(to_delete)))
                + ")"
            ).format(Identifier(SPLITGRAPH_META_SCHEMA), Identifier("object_cache_status")),
            to_delete,
        )

    def download_objects(
        self,
        source: Optional["ObjectManager"],
        objects_to_fetch: List[str],
        object_locations: List[Tuple[str, str, str]],
    ) -> List[str]:
        """
        Fetches the required objects from the remote and stores them locally.
        Does nothing for objects that already exist.

        :param source: Remote ObjectManager. If None, will only try to download objects from the external location.
        :param objects_to_fetch: List of object IDs to download.
        :param object_locations: List of custom object locations, encoded as tuples (object_id, object_url, protocol).
        """

        existing_objects = self.get_downloaded_objects(limit_to=objects_to_fetch)
        logging.debug("Need to fetch %r, already exist: %r", objects_to_fetch, existing_objects)
        objects_to_fetch = list({o for o in objects_to_fetch if o not in existing_objects})
        if not objects_to_fetch:
            logging.info("No new objects to download.")
            return []

        total_size = sum(o.size for o in self.get_object_meta(objects_to_fetch).values())
        logging.info(
            "Fetching %s, total size %s",
            pluralise("object", len(objects_to_fetch)),
            pretty_size(total_size),
        )

        # We don't actually seem to pass extra handler parameters when downloading objects since
        # we can have multiple handlers in this batch.
        external_objects = _fetch_external_objects(
            self.object_engine,
            source.metadata_engine if source else self.metadata_engine,
            object_locations,
            {},
        )

        remaining_objects_to_fetch = [o for o in objects_to_fetch if o not in external_objects]
        if not remaining_objects_to_fetch or not source:
            return external_objects

        try:
            remote_objects = self.object_engine.download_objects(
                remaining_objects_to_fetch, source.object_engine
            )
            return external_objects + remote_objects
        except IncompleteObjectDownloadError as e:
            raise IncompleteObjectDownloadError(
                reason=e.reason, successful_objects=external_objects + e.successful_objects
            )

    def upload_objects(
        self,
        target: "ObjectManager",
        objects_to_push: List[str],
        handler: str = "DB",
        handler_params: Optional[Dict[Any, Any]] = None,
    ) -> Sequence[Tuple[str, Optional[str]]]:
        """
        Uploads physical objects to the remote or some other external location.

        :param target: Target ObjectManager
        :param objects_to_push: List of object IDs to upload.
        :param handler: Name of the handler to use to upload objects. Use `DB` to push them to the remote, `FILE`
            to store them in a directory that can be accessed from the client and `HTTP` to upload them to HTTP.
        :param handler_params: For `HTTP`, a dictionary `{"username": username, "password", password}`. For `FILE`,
            a dictionary `{"path": path}` specifying the directory where the objects shall be saved.
        :return: A list of (object_id, url) that specifies all objects were uploaded (skipping objects that
            already exist on the remote).
        """
        if handler_params is None:
            handler_params = {}

        if not objects_to_push:
            logging.info("No objects to upload.")
            return []
        total_size = sum(o.size for o in self.get_object_meta(objects_to_push).values())
        logging.info(
            "Uploading %s, total size %s",
            pluralise("object", len(objects_to_push)),
            pretty_size(total_size),
        )

        if handler == "DB":
            self.object_engine.upload_objects(objects_to_push, target.object_engine)
            return [(o, None) for o in objects_to_push]

        external_handler = get_external_object_handler(handler, handler_params)

        with switch_engine(self.object_engine):
            return external_handler.upload_objects(objects_to_push, target.metadata_engine)

    def cleanup(self) -> List[str]:
        """
        Deletes all objects in the object_tree not required by any current repository, including their dependencies and
        their remote locations. Also deletes all objects not registered in the object_tree.
        """
        deleted_objects = self.cleanup_metadata()
        registered_objects = self.get_all_objects()

        # Delete unneeded/dangling objects from the cache status table
        query = SQL("DELETE FROM {}.object_cache_status").format(Identifier(SPLITGRAPH_META_SCHEMA))
        if registered_objects:
            query += SQL(" WHERE object_id != ALL(%s)")
            self.object_engine.run_chunked_sql(query, (registered_objects,), chunk_position=0)
        else:
            self.object_engine.run_sql(query)

        # Go through the physical objects and delete them as well
        # This is slightly dirty, but since the info about the objects
        # was deleted on rm, we just say that anything in splitgraph_meta
        # that's not a system table is fair game.
        tables_in_meta = {
            c
            for c in self.object_engine.get_all_tables(SPLITGRAPH_META_SCHEMA)
            if c not in META_TABLES
        }
        tables_in_meta.update(self.get_downloaded_objects())

        to_delete = [
            t for t in tables_in_meta if t not in registered_objects or t in deleted_objects
        ]
        self.delete_objects(to_delete)

        # Recalculate the object cache occupancy
        self.object_engine.run_sql(
            SQL("UPDATE {}.object_cache_occupancy SET total_size = %s").format(
                Identifier(SPLITGRAPH_META_SCHEMA)
            ),
            (self._recalculate_cache_occupancy(),),
        )
        return deleted_objects


def _fetch_external_objects(
    engine: "PostgresEngine",
    source_engine: "PsycopgEngine",
    object_locations: List[Tuple[str, str, str]],
    handler_params: Dict[str, Any],
) -> List[str]:
    non_remote_objects = []
    non_remote_by_method: DefaultDict[str, List[Tuple[str, str]]] = defaultdict(list)
    for object_id, object_url, protocol in object_locations:
        non_remote_by_method[protocol].append((object_id, object_url))
        non_remote_objects.append(object_id)

    successful: List[str] = []
    if non_remote_objects:
        for method, objects in non_remote_by_method.items():
            handler = get_external_object_handler(method, handler_params)
            # In case we're calling this from inside the FDW
            with switch_engine(engine):
                successful.extend(handler.download_objects(objects, source_engine))
    return successful



"""S3 registry-side routines called from the Python stored procedure
that are aware of the actual S3 access creds and generate pre-signed
URLs to upload/download objects."""
from datetime import timedelta
from typing import List

from minio.api import Minio
from minio.deleteobjects import DeleteObject

from splitgraph.config import CONFIG

S3_HOST = CONFIG["SG_S3_HOST"]
S3_PORT = CONFIG["SG_S3_PORT"]
S3_SECRET_KEY = CONFIG["SG_S3_PWD"]
S3_BUCKET = CONFIG["SG_S3_BUCKET"]
S3_ACCESS_KEY = CONFIG["SG_S3_KEY"]
S3_SECURE = CONFIG["SG_S3_SECURE"] == "true"

MINIO = Minio(
    "%s:%s" % (S3_HOST, S3_PORT),
    access_key=S3_ACCESS_KEY,
    secret_key=S3_SECRET_KEY,
    secure=S3_SECURE,
)

_EXP = timedelta(seconds=60)


def get_object_upload_urls(s3_host: str, object_ids: List[str]) -> List[List[str]]:
    """
    Return a list of pre-signed URLs that each part of an object can be downloaded from.

    :param s3_host: S3 host that the objects are stored on
    :param object_ids: List of object IDs
    :return: A list of lists [(object URL, object footer URL, object schema URL)]
    """
    # Currently s3_host is ignored: this is to future-proof for choosing between multiple
    # S3 providers by the client but at this point the registry knows exactly where the
    # object is located.
    return [
        [
            MINIO.presigned_put_object(
                bucket_name=S3_BUCKET, object_name=object_id + suffix, expires=_EXP
            )
            for suffix in ("", ".footer", ".schema")
        ]
        for object_id in object_ids
    ]


def get_object_download_urls(s3_host: str, object_ids: List[str]) -> List[List[str]]:
    """
    Return a list of pre-signed URLs that each part of an object can be downloaded from.

    :param s3_host: S3 host that the objects are stored on
    :param object_ids: List of object IDs
    :return: A list of lists [(object URL, object footer URL, object schema URL)]
    """
    return [
        [
            MINIO.presigned_get_object(
                bucket_name=S3_BUCKET, object_name=object_id + suffix, expires=_EXP
            )
            for suffix in ("", ".footer", ".schema")
        ]
        for object_id in object_ids
    ]


def delete_objects(client: Minio, object_ids: List[str]) -> None:
    """
    Delete objects stored in Minio

    :param client: Minio client
    :param object_ids: List of Splitgraph object IDs to delete
    """

    # Expand the list of objects into actual files we store in Minio
    all_object_ids = [o + suffix for o in object_ids for suffix in ("", ".schema", ".footer")]
    list(client.remove_objects(S3_BUCKET, map(DeleteObject, all_object_ids)))


def list_objects(client: Minio) -> List[str]:
    """
    List objects stored in Minio

    :param client: Minio client
    :return: List of Splitgraph object IDs
    """

    return [
        o.object_name
        for o in client.list_objects(bucket_name=S3_BUCKET)
        if not o.object_name.endswith(".footer") and not o.object_name.endswith(".schema")
    ]



"""
sgr commands related to mounting databases via Postgres FDW
"""

from typing import Tuple, Type

import click
from click.core import Command

from splitgraph.commandline.common import JsonType
from splitgraph.hooks.data_source import get_data_source, get_data_sources
from splitgraph.hooks.data_source.fdw import ForeignDataWrapperDataSource
from splitgraph.hooks.mount_handlers import mount


@click.group(name="mount")
def mount_c():
    """
    Mount foreign databases as Postgres schemas.

    Uses the Postgres FDW interface to create a local Postgres schema with foreign tables that map
    to tables in other databases.

    See a given mount handler's documentation for handler-specific parameters.
    """


def _generate_handler_help(handler: Type[ForeignDataWrapperDataSource]) -> Tuple[str, str]:
    """
    Extract the long description and the parameters from a handler's class
    """
    handler_help = (
        handler.commandline_help or handler.get_name() + "\n\n" + handler.get_description()
    )

    if handler.commandline_kwargs_help:
        handler_kwargs_help = (
            "JSON-encoded dictionary or @filename.json with handler options:\n\n"
            + handler.commandline_kwargs_help
        )
    else:
        handler_kwargs_help = "JSON-encoded dictionary or @filename.json with handler options"

    return handler_help, handler_kwargs_help


def _make_mount_handler_command(
    handler_name: str, handler: Type[ForeignDataWrapperDataSource]
) -> Command:
    """Turn the mount handler function into a Click subcommand
    with help text and kwarg/connection string passing"""

    help_text, handler_options_help = _generate_handler_help(handler)

    params = [
        click.Argument(["schema"]),
        click.Option(
            ["--connection", "-c"],
            help="Connection string in the form username:password@server:port",
        ),
        click.Option(
            ["--handler-options", "-o"], help=handler_options_help, default="{}", type=JsonType()
        ),
    ]
    from splitgraph.core.output import conn_string_to_dict

    def _callback(schema, connection, handler_options):
        handler_options.update(conn_string_to_dict(connection))
        mount(schema, mount_handler=handler_name, handler_kwargs=handler_options)

    cmd = click.Command(handler_name, params=params, callback=_callback, help=help_text)
    return cmd


# Register all current mount handlers and turn them into Click subcommands.
for _handler_name in get_data_sources():
    _handler = get_data_source(_handler_name)
    if not _handler.supports_mount:
        continue
    assert issubclass(_handler, ForeignDataWrapperDataSource)
    mount_c.add_command(_make_mount_handler_command(_handler_name, _handler))



import base64
import itertools
import os
import random
import string
from io import StringIO
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Tuple

import ruamel.yaml
from pydantic import BaseModel, Field
from ruamel.yaml import CommentedMap as CM
from ruamel.yaml import CommentedSeq as CS

from splitgraph.cloud import GQLAPIClient, Plugin
from splitgraph.cloud.project.dbt import (
    generate_dbt_plugin_params,
    generate_dbt_project,
)
from splitgraph.cloud.project.github_actions import generate_workflow
from splitgraph.cloud.project.templates import README_TEMPLATE, SPLITGRAPH_YML_TEMPLATE


def get_comment(jsonschema_object: Any) -> str:
    """
    Get a YAML comment to be attached to a JSONSchema property.
    """
    result = []
    if "title" in jsonschema_object:
        result.append(jsonschema_object["title"])
    if "description" in jsonschema_object:
        result.append(jsonschema_object["description"].strip().replace("\n", " "))
    if "enum" in jsonschema_object:
        enum = jsonschema_object["enum"]
        if len(enum) == 1:
            result.append("Constant")
        else:
            result.append("One of " + ", ".join(enum))
    if "const" in jsonschema_object and "Constant" not in result:
        result.append("Constant")
    if "oneOf" in jsonschema_object:
        result.append("Choose one of:")

    return ". ".join(result).strip()


_SCALARS = ("string", "integer", "boolean")


def jsonschema_object_to_example(obj: Any, type_override=None) -> Any:
    """
    Get an example value for a JSONSchema property
    """
    obj_type = type_override or obj["type"]
    if obj_type in _SCALARS:
        if "examples" in obj:
            return obj["examples"][0]
        elif "default" in obj:
            return obj["default"]
        elif "enum" in obj:
            return obj["enum"][0]
        elif "const" in obj:
            return obj["const"]
        else:
            return {"string": "", "integer": 0, "boolean": False}.get(obj_type)
    elif obj_type == "array":
        # For simple lists of scalars, don't drill down and just return an empty list
        if obj["items"]["type"] in _SCALARS:
            seq = CS([])
        else:
            example = jsonschema_object_to_example(obj["items"])
            seq = CS([example])
        comment = get_comment(obj["items"])
        if comment:
            seq.yaml_set_start_comment(comment)
        return seq
    elif obj_type == "object":
        if "properties" in obj:
            return _get_object_example(obj)
        elif "oneOf" in obj:
            return _get_oneof_example(obj)
        return {}


def _get_oneof_example(obj: Dict[str, Any]) -> CS:
    result = CS()
    for o in obj["oneOf"]:
        result.append(jsonschema_object_to_example(o, "object"))
        comment = get_comment(o)
        if comment:
            result.yaml_add_eol_comment(comment, key=len(result) - 1)
    return result


def _get_object_example(obj: Dict[str, Any]) -> CM:
    properties = obj["properties"]
    required = obj.get("required", [])
    required_items = []
    not_required_items = []
    result = CM()
    for p in properties:
        example = jsonschema_object_to_example(properties[p])
        comment = get_comment(properties[p])
        if p in required:
            if comment:
                comment = "REQUIRED. " + comment
            else:
                comment = "REQUIRED"
            required_items.append((p, example, comment))
        else:
            not_required_items.append((p, example, comment))
    for item, example, comment in itertools.chain(required_items, not_required_items):
        result[item] = example
        if comment:
            result.yaml_add_eol_comment(comment, key=item)
    return result


def stub_plugin(plugin: Plugin, namespace: str, repository: str, is_live: bool = False) -> CM:
    """
    Generate a splitgraph.yml file based on a plugin's JSONSchemas.
    """
    yml = ruamel.yaml.YAML()
    repositories_yaml = (
        SPLITGRAPH_YML_TEMPLATE.replace("CREDENTIAL_NAME", plugin.plugin_name)
        .replace("NAMESPACE", namespace)
        .replace("REPOSITORY", repository)
        .replace("PLUGIN_NAME", plugin.plugin_name)
    )
    ruamel_dict = yml.load(StringIO(repositories_yaml))
    assert isinstance(ruamel_dict, CM)

    ruamel_dict["credentials"][plugin.plugin_name]["data"] = jsonschema_object_to_example(
        plugin.credentials_schema
    )

    ruamel_dict["repositories"][0]["external"]["params"] = jsonschema_object_to_example(
        plugin.params_schema
    )
    ruamel_dict["repositories"][0]["external"]["tables"]["sample_table"][
        "options"
    ] = jsonschema_object_to_example(plugin.table_params_schema)
    ruamel_dict["repositories"][0]["external"]["is_live"] = is_live

    return ruamel_dict


def _get_seed_uid() -> str:
    return "".join(
        random.choice(string.ascii_lowercase + string.ascii_uppercase + string.digits)
        for _ in range(10)
    )


class ProjectSeed(BaseModel):
    """
    Contains all information required to generate a Splitgraph project + optionally
    a dbt model for GitHub Actions
    """

    seed_uid: str = Field(default_factory=_get_seed_uid)
    namespace: str
    plugins: List[str]
    include_dbt: bool = False

    def encode(self) -> str:
        return base64.b64encode(self.json(separators=(",", ":")).encode()).decode()

    @classmethod
    def decode(cls, encoded: str) -> "ProjectSeed":
        return ProjectSeed.parse_raw(base64.b64decode(encoded.encode()))


def generate_project(
    api_client: GQLAPIClient, seed: str, basedir: Path, github_repo: Optional[str] = None
) -> None:
    all_plugins = {p.plugin_name: p for p in api_client.get_all_plugins(seed)}

    decoded_seed = ProjectSeed.decode(seed)
    credentials, repositories, repository_info = generate_splitgraph_yml(all_plugins, decoded_seed)

    yml = ruamel.yaml.YAML()
    with open(os.path.join(basedir, "splitgraph.credentials.yml"), "w") as f:
        yml.dump(credentials, f)

    with open(os.path.join(basedir, "splitgraph.yml"), "w") as f:
        yml.dump(repositories, f)

    # Generate the dbt project
    if decoded_seed.include_dbt:
        dbt_repo, _, is_dbt = repository_info[-1]
        assert is_dbt
        dbt_sources = [r for r, _, is_dbt in repository_info if not is_dbt]
        dependencies = {dbt_repo: dbt_sources}
        generate_dbt_project(dbt_sources, basedir)
    else:
        dependencies = {}

    # Generate the Github workflow file
    github_root = os.path.join(basedir, ".github/workflows")
    os.makedirs(github_root, exist_ok=True)

    with open(os.path.join(github_root, "build.yml"), "w") as f:
        yml.dump(generate_workflow(repository_info, dependencies), f)

    # Generate the README
    with open(os.path.join(basedir, "README.md"), "w") as f:
        template = README_TEMPLATE
        # Add the GitHub repo to some places that might need it, e.g. URLs to the
        # repo settings page
        if github_repo:
            template = template.replace("$GITHUB_REPO", github_repo)
        f.write(template)


def generate_splitgraph_yml(
    all_plugins: Mapping[str, Plugin], seed: ProjectSeed
) -> Tuple[CM, CM, List[Tuple[str, bool, bool]]]:
    repository_info: List[Tuple[str, bool, bool]] = []
    repository_names: List[str] = []

    credentials = CM({"credentials": CM({})})
    repositories = CM({"repositories": CS()})
    for plugin_name in seed.plugins:
        plugin = all_plugins[plugin_name]
        stub = stub_plugin(
            plugin, namespace=seed.namespace, repository=plugin_name, is_live=plugin.supports_mount
        )

        target_repo = f"{seed.namespace}/{plugin_name}"
        repository_info.append((target_repo, plugin.supports_mount, False))
        repository_names.append(target_repo)
        credentials["credentials"].update(stub["credentials"])
        repositories["repositories"].extend(stub["repositories"])

    if seed.include_dbt:
        dbt_params, dbt_credentials = generate_dbt_plugin_params(repository_names)
        dbt_repo = f"{seed.namespace}/dbt-sample"
        credential_name = "dbt-sample"
        repositories["repositories"].append(
            CM(
                {
                    "namespace": seed.namespace,
                    "repository": "dbt-sample",
                    "external": CM(
                        {
                            "plugin": "dbt",
                            "credential": credential_name,
                            "params": dbt_params,
                            "is_live": False,
                            "tables": CM({}),
                        }
                    ),
                    "metadata": CM(
                        {
                            "description": "Sample dbt model",
                            "readme": {
                                "text": "## Sample dbt model\n\n"
                                "This is an autogenerated model referencing data from:\n\n"
                                + "\n".join(f"  * [/{r}](/{r})" for r in repository_names)
                            },
                        }
                    ),
                }
            )
        )
        credentials["credentials"][credential_name] = CM({"plugin": "dbt", "data": dbt_credentials})
        repository_info.append((dbt_repo, False, True))

    return (
        credentials,
        repositories,
        repository_info,
    )



"""
sgr commands related to sharing and downloading images.
"""

import sys
from typing import Optional

import click

from splitgraph.commandline.common import ImageType, JsonType, RepositoryType
from splitgraph.config import CONFIG, REMOTES
from splitgraph.config.config import get_from_subsection


@click.command(name="pull")
@click.argument("repository_or_image", type=ImageType(repository_exists=True, default=None))
@click.option(
    "-d",
    "--download-all",
    is_flag=True,
    help="Download all objects immediately instead of on checkout.",
)
@click.option(
    "-f",
    "--overwrite-object-meta",
    help="Overwrite metadata for existing objects",
    is_flag=True,
)
@click.option("-t", "--tags", help="Overwrite tags for pulled image/repo", is_flag=True)
def pull_c(repository_or_image, download_all, overwrite_object_meta, tags):
    """
    Pull changes / download a single image.
    """
    repository, image = repository_or_image
    repository.pull(
        download_all,
        single_image=image,
        overwrite_objects=overwrite_object_meta,
        overwrite_tags=tags,
    )


@click.command(name="clone")
@click.argument("remote_repository_or_image", type=ImageType(default=None))
@click.argument("local_repository", required=False, type=RepositoryType())
@click.option("-r", "--remote", help="Name of the remote engine")
@click.option(
    "-d",
    "--download-all",
    help="Download all objects immediately instead of on checkout.",
    default=False,
    is_flag=True,
)
@click.option(
    "-f",
    "--overwrite-object-meta",
    help="Overwrite metadata for existing objects",
    is_flag=True,
)
@click.option("-t", "--tags", help="Overwrite tags for pulled image/repo", is_flag=True)
def clone_c(
    remote_repository_or_image, local_repository, remote, download_all, overwrite_object_meta, tags
):
    """
    Clone a remote Splitgraph repository/image into a local one.

    The lookup path for the repository is governed by the ``SG_REPO_LOOKUP`` and ``SG_REPO_LOOKUP_OVERRIDE``
    config parameters and can be overridden by the command line ``--remote`` option.
    """
    from splitgraph.core.repository import Repository, clone
    from splitgraph.engine import get_engine

    remote_repository, image = remote_repository_or_image

    # If the user passed in a remote, we can inject that into the repository spec.
    # Otherwise, we have to turn the repository into a string and let clone() look up the
    # actual engine the repository lives on.
    if remote:
        remote_repository = Repository.from_template(remote_repository, engine=get_engine(remote))
    else:
        remote_repository = remote_repository.to_schema()

    clone(
        remote_repository,
        local_repository=local_repository,
        download_all=download_all,
        single_image=image,
        overwrite_objects=overwrite_object_meta,
        overwrite_tags=tags,
    )


@click.command(name="push")
@click.argument("repository_or_image", type=ImageType(default=None))
@click.argument("remote_repository", required=False, type=RepositoryType())
@click.option(
    "-r",
    "--remote",
    help="Name of the remote engine",
    type=click.Choice(REMOTES),
    default=None,
)
@click.option("-h", "--upload-handler", help="Upload handler", default="S3")
@click.option(
    "-o",
    "--upload-handler-options",
    help="Upload handler parameters",
    default="{}",
    type=JsonType(),
)
@click.option(
    "-f",
    "--overwrite-object-meta",
    help="Overwrite metadata for existing remote objects",
    is_flag=True,
)
@click.option("-t", "--tags", help="Overwrite tags for pushed image/repo", is_flag=True)
@click.option(
    "-u",
    "--reupload-objects",
    help="Reupload objects that metadata is pushed out for",
    is_flag=True,
)
def push_c(
    repository_or_image,
    remote_repository,
    remote,
    upload_handler,
    upload_handler_options,
    overwrite_object_meta,
    reupload_objects,
    tags,
):
    """
    Push images to the Splitgraph registry or another engine.

    If an image is not specified (e.g. `sgr push noaa/climate`, this will push all new images.

    By default, the repository will be pushed to a repository with the same name in the user's namespace
    (SG_NAMESPACE configuration value which defaults to the username).

    If there's a single engine registered in the config (e.g. data.splitgraph.com), it shall be the default
    destination.

    If an upstream repository/engine has been configured for this engine with `sgr upstream`,
    it will be used instead.

    Finally, if `remote_repository` or `--remote` are passed, they will take precedence.

    The actual objects will be uploaded to S3 using Minio. When pushing to another engine,
    you can choose to upload them directly by passing `--handler DB`.
    """
    repository, image = repository_or_image

    remote_repository = _determine_push_target(repository, remote_repository, remote)

    click.echo(
        "Pushing %s to %s on remote %s"
        % (repository, remote_repository, remote_repository.engine.name)
    )

    repository.push(
        remote_repository,
        handler=upload_handler,
        handler_options=upload_handler_options,
        overwrite_objects=overwrite_object_meta,
        reupload_objects=reupload_objects,
        overwrite_tags=tags,
        single_image=image,
    )


def _determine_push_target(repository, remote_repository, remote):
    """
    Create the remote Repository object we're pushing to based on all the
    parameters we've been passed

    The reason for this behaviour is to streamline out-of-the-box Splitgraph setups where
    data.splitgraph.com is the only registered engine. In that case:

     * sgr push repo: will push to myself/repo on data.splitgraph.com with S3 uploading (user's namespace).
     * sgr push noaa/climate: will push to myself/climate
     * sgr push noaa/climate noaa/climate: will explicitly push to noaa/climate (assuming the user can write
       to that repository).

    If the user registers another registry at splitgraph.mycompany.com, then they will be able to do:

    * sgr push noaa/climate -r splitgraph.mycompany.com: will push to noaa/climate

    :param repository: Local Repository, required.
    :param remote_repository: remote Repository (without the remote engine), optional.
    :param remote: Name of the remote engine/registry, optional.
    :return:
    """
    from splitgraph.core.repository import Repository
    from splitgraph.engine import get_engine

    if remote_repository and remote:
        remote_repository = Repository.from_template(remote_repository, engine=get_engine(remote))
    elif remote:
        remote_repository = _make_push_target(repository, remote)
    elif remote_repository:
        remote_repository = Repository.from_template(
            remote_repository, engine=get_engine(_get_default_remote())
        )
    else:
        remote_repository = repository.upstream or _make_push_target(
            repository, _get_default_remote()
        )
    return remote_repository


def _get_default_remote():
    if len(REMOTES) != 1:
        raise click.UsageError(
            "More than one remote registered, specify a remote explicitly with -r!"
        )
    remote = REMOTES[0]
    return remote


def _make_push_target(repository, remote):
    from splitgraph.core.repository import Repository
    from splitgraph.engine import get_engine

    namespace: Optional[str]
    try:
        namespace = get_from_subsection(CONFIG, "remotes", remote, "SG_NAMESPACE")
    except KeyError:
        namespace = None
    remote_repository = Repository.from_template(
        repository, namespace=namespace, engine=get_engine(remote)
    )
    return remote_repository


@click.command(name="upstream")
@click.argument("repository", type=RepositoryType(exists=True))
@click.option(
    "-s",
    "--set",
    "set_to",
    help="Set the upstream to a engine alias + repository",
    type=(str, RepositoryType()),
    default=("", None),
)
@click.option("-r", "--reset", help="Delete the upstream", is_flag=True, default=False)
def upstream_c(repository, set_to, reset):
    """
    Get or set the upstream for a repository.

    This shows the default repository used for pushes and pulls as well as allows to change it to a different
    remote engine and repository.

    The remote engine alias must exist in the config file.

    Examples:

    ``sgr upstream my/repo --set splitgraph.com username/repo``

    Sets the upstream for ``my/repo`` to ``username/repo`` existing on the ``splitgraph.com`` engine

    ``sgr upstream my/repo --reset``

    Removes the upstream for ``my/repo``.

    ``sgr upstream my/repo``

    Shows the current upstream for ``my/repo``.
    """
    from splitgraph.core.repository import Repository
    from splitgraph.engine import get_engine

    # surely there's a better way of finding out whether --set isn't specified
    if set_to != ("", None) and reset:
        raise click.BadParameter("Only one of --set and --reset can be specified!")

    if reset:
        if repository.upstream:
            del repository.upstream
            click.echo("Deleted upstream for %s." % repository.to_schema())
        else:
            click.echo("%s has no upstream to delete!" % repository.to_schema())
            sys.exit(1)
        return

    if set_to == ("", None):
        upstream = repository.upstream
        if upstream:
            click.echo(
                "%s is tracking %s:%s."
                % (repository.to_schema(), upstream.engine.name, upstream.to_schema())
            )
        else:
            click.echo("%s has no upstream." % repository.to_schema())
    else:
        engine, remote_repo = set_to
        try:
            remote_repo = Repository.from_template(remote_repo, engine=get_engine(engine))
        except KeyError:
            click.echo("Remote engine '%s' does not exist in the configuration file!" % engine)
            sys.exit(1)
        repository.upstream = remote_repo
        click.echo(
            "%s set to track %s:%s." % (repository.to_schema(), engine, remote_repo.to_schema())
        )



"""Core Splitgraph functionality: versioning and sharing tables.

The main point of interaction with the Splitgraph API is a :class:`splitgraph.core.repository.Repository` object
representing a local or a remote Splitgraph repository. Repositories can be created using one of the
following methods:

  * Directly by invoking `Repository(namespace, name, engine)` where `engine` is the engine that the repository
    belongs to (that can be gotten with `get_engine(engine_name)`. If the created repository doesn't actually exist
    on the engine, it must first be initialized with `repository.init()`.
  * By using :func:`splitgraph.core.engine.lookup_repository` which will search for the repository on the current
    lookup path.
"""



from typing import Any, Optional

from .keys import DEFAULTS


def get_default_config_value(key: str, default_return: Optional[str] = None) -> Any:
    """Get the hard-coded default value of a config key.
    Otherwise return default_return.
    """

    return DEFAULTS.get(key, default_return)



"""
Routines for managing Splitgraph engines, including looking up repositories and managing objects.
"""
import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from psycopg2.sql import SQL, Identifier

from splitgraph.config import CONFIG, SPLITGRAPH_API_SCHEMA, get_singleton
from splitgraph.engine import ResultShape, get_engine
from splitgraph.exceptions import RepositoryNotFoundError

from .sql import select

if TYPE_CHECKING:
    from splitgraph.core.image import Image
    from splitgraph.core.repository import Repository
    from splitgraph.engine.postgres.engine import PostgresEngine


def _parse_paths_overrides(
    lookup_path: str, override_path: str
) -> Tuple[List[str], Dict[str, str]]:
    return (
        lookup_path.split(",") if lookup_path else [],
        {r[: r.index(":")]: r[r.index(":") + 1 :] for r in override_path.split(",")}
        if override_path
        else {},
    )


# Parse and set these on import. If we ever need to be able to reread the config on the fly, these have to be
# recalculated.
_LOOKUP_PATH, _LOOKUP_PATH_OVERRIDE = _parse_paths_overrides(
    get_singleton(CONFIG, "SG_REPO_LOOKUP"), get_singleton(CONFIG, "SG_REPO_LOOKUP_OVERRIDE")
)


def init_engine(skip_object_handling: bool = False) -> None:  # pragma: no cover
    # Method exercised in test_commandline.test_init_new_db but in
    # an external process
    """
    Initializes the engine by:

        * performing any required engine-custom initialization
        * creating the metadata tables

    :param skip_object_handling: If True, skips installing routines related to
        object handling and checkouts (like audit triggers and CStore management).
    """
    # Initialize the engine
    engine = get_engine()
    engine.initialize(skip_object_handling=skip_object_handling)
    engine.commit()
    logging.info("Engine %r initialized.", engine)


def repository_exists(repository: "Repository") -> bool:
    """
    Checks if a repository exists on the engine.

    :param repository: Repository object
    """
    return (
        repository.engine.run_sql(
            SQL("SELECT 1 FROM {}.get_images(%s,%s)").format(Identifier(SPLITGRAPH_API_SCHEMA)),
            (repository.namespace, repository.repository),
            return_shape=ResultShape.ONE_ONE,
        )
        is not None
    )


def lookup_repository(name: str, include_local: bool = False) -> "Repository":
    """
    Queries the SG engines on the lookup path to locate one hosting the given repository.

    :param name: Repository name
    :param include_local: If True, also queries the local engine

    :return: Local or remote Repository object
    """
    from splitgraph.core.repository import Repository

    template = Repository.from_schema(name)

    if name in _LOOKUP_PATH_OVERRIDE:
        return Repository(
            template.namespace, template.repository, get_engine(_LOOKUP_PATH_OVERRIDE[name])
        )

    # Currently just check if the schema with that name exists on the remote.
    if include_local and repository_exists(template):
        return template

    for engine in _LOOKUP_PATH:
        candidate = Repository(template.namespace, template.repository, get_engine(engine))
        if repository_exists(candidate):
            return candidate
        candidate.engine.close()

    raise RepositoryNotFoundError("Unknown repository %s!" % name)


def get_current_repositories(
    engine: "PostgresEngine",
) -> List[Tuple["Repository", Optional["Image"]]]:
    """
    Lists all repositories currently in the engine.

    :param engine: Engine
    :return: List of (Repository object, current HEAD image)
    """
    from splitgraph.core.repository import Repository

    all_repositories = [
        Repository(n, r, engine)
        for n, r in engine.run_sql(select("images", "DISTINCT namespace,repository"))
    ]
    return [(r, r.head) for r in all_repositories]



from typing import Any

from ruamel.yaml import YAML


def safe_load(stream) -> Any:
    yaml = YAML(typ="safe")
    return yaml.load(stream)


def safe_dump(obj: Any, stream, **kwargs) -> None:
    yaml = YAML(typ="safe")
    yaml.default_flow_style = False
    yaml.dump(obj, stream, **kwargs)



"""
Internal functions for parsing Splitfiles.
"""

import re
import shlex
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Tuple, TypeVar, Union

from parsimonious import Grammar
from parsimonious.nodes import Node, RegexNode

from splitgraph.exceptions import SplitfileError

if TYPE_CHECKING:
    from splitgraph.core.repository import Repository


SPLITFILE_GRAMMAR = Grammar(
    r"""
    commands = space? command space_nn? (newline space? command space_nn?)* space?
    command = comment / import / from / sql_file / sql / custom
    comment = space "#" non_newline?
    from = "FROM" space ("EMPTY" / repo_source) (space "AS" space repository)?
    import = "FROM" space source space "IMPORT" space tables
    sql_file = "SQL" space "FILE" space non_newline
    sql = "SQL" space (("{" non_curly_brace "}") / non_newline)
    custom = identifier space non_newline

    table = ((table_name / table_query) space "AS" space table_alias) / table_name

    table_query = "{" non_curly_brace "}"
    tables = "ALL" / (table (space "," space table)*)
    source = mount_source / repo_source
    repo_source = repository (":" tag_or_hash)?
    mount_source = "MOUNT" space handler space no_db_conn_string space handler_options

    image_hash = ~"[0-9a-f]*"i

    # TBH these ones that map to "identifier" aren't really necessary since parsimonious calls those nodes
    # "identifier" anyway. This is so that the grammar is slightly more readable.

    handler = identifier
    repository = ~"[_a-zA-Z0-9-/]+"
    table_name = identifier
    table_alias = identifier
    tag_or_hash = ~"\S+"
    handler_options = "'" non_single_quote "'"

    newline = ~"\n+"
    non_newline = ~"[^\n]+"

    # I've no idea why we need so many slashes here. The purpose of this regex is to consume anything
    # that's not a closing curly brace or \} (an escaped curly brace).
    non_curly_brace = ~"(\\\\}|[^}])*"

    # Yeah, six slashes should be about enough to capture \'
    non_single_quote = ~"(\\\\\\'|[^'])*"

    no_db_conn_string = ~"((\S+):(\S+)@(.+):(\d+))?"
    identifier = ~"[_a-zA-Z0-9-]+"
    space = ~"\s*"
    space_nn = ~"[ \t]*"
"""
)


def preprocess(commands: str, params: Optional[Dict[str, str]] = None) -> str:
    """
    Preprocesses a Splitfile, performing parameter substitution (`${PARAM}` gets replaced with `params['PARAM']`).
    Also removes escaped newlines.
    Raises an error if not all parameters are specified.

    :param commands: A string with the raw Splitfile
    :param params: Dictionary of parameters to substitute
    :returns The preprocessed Splitfile.
    """
    # Also replaces all $PARAM in the splitfile text with the params in the dictionary.
    if params is None:
        params = {}
    commands = commands.replace("\\\n", "")
    for param, value in params.items():
        # Regex fun: if the replacement is '\1' + substitution (so that we put back the previously-consumed
        # possibly-escape character) and the substitution begins with a number (like an IP address),
        # then it gets treated as a match group (say, \11) which fails silently and adds weird gibberish
        # to the result.
        commands = re.sub(
            r"([^\\])\${" + re.escape(param) + "}",
            r"\g<1>" + str(value),
            commands,
            flags=re.MULTILINE,
        )
    # Search for any unreplaced $-parameters
    unreplaced = set(re.findall(r"[^\\](\${\S+})", commands, flags=re.MULTILINE))
    if unreplaced:
        raise SplitfileError("Unknown values for parameters " + ", ".join(unreplaced) + "!")
    # Finally, replace the escaped $
    return commands.replace("\\$", "$")


def parse_commands(commands: str, params: Optional[Dict[str, str]] = None) -> List[Node]:
    """Unpacks the parse tree into a list of command nodes."""
    if params is None:
        params = {}
    commands = preprocess(commands, params)
    parse_tree = SPLITFILE_GRAMMAR.parse(commands)
    return [
        n.children[0]
        for n in extract_nodes(parse_tree, ["command"])
        if n.children[0].expr_name != "comment"
    ]


def extract_nodes(node: Node, types: List[str]) -> List[Node]:
    """Crawls the parse tree and only extracts nodes of given types. Doesn't crawl further down if it reaches a
    sought type."""
    if node.expr_name in types:
        return [node]
    result = []
    for child in node.children:
        result.extend(extract_nodes(child, types))
    return result


def get_first_or_none(
    node_list: Union[List[RegexNode], List[Node]], node_type: str
) -> Optional[Node]:
    """Gets the first node of type node_type from node_list, returns None if it doesn't exist."""
    for node in node_list:
        if node.expr_name == node_type:
            return node
    return None


def _parse_table_alias(table_node: Node) -> Tuple[str, str, bool]:
    """Extracts the table name (or a query forming the table) and its alias from the parse tree."""
    table_name_alias = extract_nodes(table_node, ["identifier", "non_curly_brace"])
    table_name = table_name_alias[0].match.group(0)
    table_is_query = table_name_alias[0].expr_name == "non_curly_brace"
    if table_is_query:
        # Unescape the closing curly brace that marked the end of the query
        table_name = table_name.replace("\\}", "}")
    if len(table_name_alias) > 1:
        table_alias = table_name_alias[1].match.group(0)
        return table_name, table_alias, table_is_query
    return table_name, table_name, table_is_query


def parse_image_spec(remote_repo_node: Node) -> Tuple["Repository", str]:
    """
    Extracts the image specification (e.g. noaa/climate:abcdef123 -> Repository('noaa', 'climate'), 'abcdef123')
    :param remote_repo_node: Parse node with the specification
    :return: Tuple of (repository object, tag or hash)
    """
    repo_nodes = extract_nodes(remote_repo_node, ["repository", "tag_or_hash"])
    # Avoid cyclic imports
    from splitgraph.core.repository import Repository

    repository = Repository.from_schema(repo_nodes[0].match.group(0))
    # See if we got given a tag / hash (the executor will try to interpret it as both).
    if len(repo_nodes) == 2:
        tag_or_hash = repo_nodes[1].match.group(0)
    else:
        tag_or_hash = "latest"
    return repository, tag_or_hash


T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")


def _transpose3(seq: Sequence[Tuple[T1, T2, T3]]) -> Tuple[List[T1], List[T2], List[T3]]:
    t1, t2, t3 = zip(*seq)
    return list(t1), list(t2), list(t3)


def extract_all_table_aliases(node: Node) -> Tuple[List[str], List[str], List[bool]]:
    """
    Extracts table names and aliases in a format suitable for passing to the `import_tables` function
    :param node: Parse node
    :return: Three lists: table names, table aliases (can be the same as table names), whether table names are
        actually SQL queries defining data to insert.
    """
    tables = extract_nodes(node, ["table"])
    if not tables:
        # No tables specified (imports all tables from the mounted db / repo)
        return [], [], []
    return _transpose3([_parse_table_alias(table) for table in tables])


def parse_custom_command(node: Node) -> Tuple[str, List[str]]:
    """Splits the parse tree node (CMD arg1 --arg2 "arg 3") into a tuple (command, args)."""
    repo_nodes = extract_nodes(node, ["identifier", "non_newline"])
    command = repo_nodes[0].match.group(0)
    # Use the sh-like lexer to split args up so that e.g. 'arg1 --arg2 "arg 3"' turns into ['arg1', '--arg2', 'arg 3']
    return command, shlex.split(repo_nodes[1].match.group(0))



import logging
import traceback
from collections import Callable
from functools import wraps
from typing import Any, Dict, Optional

from psycopg2.sql import SQL, Identifier

from splitgraph.core.repository import Repository
from splitgraph.core.types import Changeset, TableSchema
from splitgraph.engine import validate_type
from splitgraph.engine.postgres.engine import PostgresEngine, get_change_key
from splitgraph.hooks.data_source.base import (
    INGESTION_STATE_SCHEMA,
    INGESTION_STATE_TABLE,
)

SingerConfig = Dict[str, Any]
SingerCatalog = Dict[str, Any]
SingerState = Dict[str, Any]


def log_exception(f):
    """Emit exceptions with full traceback instead of just the error text"""

    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception:
            logging.error(traceback.format_exc())
            raise

    return wrapped


def rollback_at_end(func: Callable) -> Callable:
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        repository = self.image.repository
        try:
            return func(self, *args, **kwargs)
        finally:
            repository.rollback_engines()

    return wrapped


def _migrate_schema(engine, table_schema, table_name, table_schema_spec, new_schema_spec):
    """Migrate the schema of a table to match the schema_spec"""

    old_cols = {c.name: c.pg_type for c in table_schema_spec}
    new_cols = {c.name: c.pg_type for c in new_schema_spec}
    for c in old_cols:
        if c not in new_cols:
            engine.run_sql(
                SQL("ALTER TABLE {}.{} DROP COLUMN {}").format(
                    Identifier(table_schema),
                    Identifier(table_name),
                    Identifier(c),
                )
            )
    for c in new_cols:
        if c not in old_cols:
            engine.run_sql(
                SQL("ALTER TABLE {}.{} ADD COLUMN {} %s" % validate_type(new_cols[c])).format(
                    Identifier(table_schema),
                    Identifier(table_name),
                    Identifier(c),
                )
            )
        elif new_cols[c] != old_cols[c]:
            engine.run_sql(
                SQL(
                    "ALTER TABLE {}.{} ALTER COLUMN {} TYPE %s" % validate_type(new_cols[c])
                ).format(
                    Identifier(table_schema),
                    Identifier(table_name),
                    Identifier(c),
                )
            )


def _make_changeset(
    engine: PostgresEngine,
    schema: str,
    table: str,
    schema_spec: TableSchema,
    upsert_condition: str = "TRUE",
) -> Changeset:
    """Build a fake changeset from the temporary table and the existing table to pass
    to the object manager (store as a Splitgraph diff)."""

    # PK -> (upserted / deleted, old row, new row)
    # We don't find out the old row here. This is because it requires a JOIN on the current
    # Splitgraph table, so if we're adding e.g. 100k rows to a 1M row table, it's going to cause big
    # performance issues. Instead, we pretend that all rows
    # have been inserted (apart from the ones that have been deleted by having the magic
    # _sdc_deleted_at column).

    # We also don't care about finding out the new row here, as the storage routine queries
    # the table directly to get those values.

    # The tradeoff is that now, when querying the table, we need to include not only fragments
    # whose index matches the query, but also all fragments that might overwrite those fragments
    # (through PK range overlap). Since we don't record old row values in this changeset's index,
    # we can no longer find if a fragment deletes some row by inspecting the index -- we need to
    # use PK ranges to find out overlapping fragments.

    change_key = [c for c, _ in get_change_key(schema_spec)]
    # Query:
    # SELECT (col_1, col_2, ...) AS pk,
    #        (custom upsert condition),
    #        {} AS old_row
    # FROM new_table n
    query = (
        SQL("SELECT ")
        + SQL(",").join(SQL("n.") + Identifier(c) for c in change_key)
        + SQL(",")
        + SQL(upsert_condition + " AS upserted FROM {}.{} n").format(
            Identifier(schema), Identifier(table)
        )
    ).as_string(engine.connection)
    result = engine.run_sql(query)
    return {tuple(row[:-1]): (row[-1], {}, {}) for row in result}


def store_ingestion_state(
    repository: Repository,
    image_hash: str,
    current_state: Optional[SingerState],
    new_state: str,
):
    # Add a table to the new image with the new state
    repository.object_engine.create_table(
        schema=None,
        table=INGESTION_STATE_TABLE,
        schema_spec=INGESTION_STATE_SCHEMA,
        temporary=True,
    )
    # NB: new_state here is a JSON-serialized string, so we don't wrap it into psycopg2.Json()
    logging.info("Writing state: %s", new_state)
    repository.object_engine.run_sql(
        SQL("INSERT INTO pg_temp.{} (timestamp, state) VALUES(now(), %s)").format(
            Identifier(INGESTION_STATE_TABLE)
        ),
        (new_state,),
    )
    object_id = repository.objects.create_base_fragment(
        "pg_temp", INGESTION_STATE_TABLE, repository.namespace, table_schema=INGESTION_STATE_SCHEMA
    )
    # If the state exists already, overwrite it; otherwise, add new state table.
    if current_state:
        repository.objects.overwrite_table(
            repository,
            image_hash,
            INGESTION_STATE_TABLE,
            INGESTION_STATE_SCHEMA,
            [object_id],
        )
    else:
        repository.objects.register_tables(
            repository,
            [(image_hash, INGESTION_STATE_TABLE, INGESTION_STATE_SCHEMA, [object_id])],
        )



"""Bloom filtering on fragments for equality queries."""
import base64
import itertools
import struct
from datetime import datetime
from hashlib import sha256
from math import ceil, exp, log
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union, cast

from psycopg2.sql import SQL, Identifier

from splitgraph.config import SPLITGRAPH_META_SCHEMA
from splitgraph.core.output import pretty_size
from splitgraph.core.types import Changeset
from splitgraph.engine.postgres.engine import SG_UD_FLAG

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PsycopgEngine


def _hash_value(value: Union[datetime, int, str, None]) -> Tuple[bytes, bytes]:
    if value is None:
        # See explanation in generate_bloom_index for why this isn't
        # a completely terrible idea.
        value = "NULL"
    return (
        sha256(str(value).encode("utf-8")).digest(),
        sha256((str(value) + "salt").encode("utf-8")).digest(),
    )


def generate_bloom_index(
    engine: "PsycopgEngine",
    object_id: str,
    changeset: Optional[Changeset],
    column: str,
    probability: Optional[float] = None,
    size: Optional[int] = None,
) -> Tuple[int, str]:
    """
    Generates a bloom filter signature for a given column and a given fragment. Bloom filters
    can answer queries asking whether an item is definitely not in a given set or possibly can be.

    The tradeoff is between the probability of a false positive (item said to be in the set when
    it actually isn't) and the size of the filter.

    Bloom filters also have an extra parameter, k, or the number of bits in the signature that
    a certain item flips. This parameter has an optimal value for a given number of distinct items
    or a probability and so isn't explicitly passed by the user.

    :param engine: Object engine the fragment is cached in.
    :param object_id: Fragment ID
    :param changeset: Optional, if specified, the old column values are included in the index.
    :param column: Column name to generate the index on.
    :param probability: Probability of a false positive. Either this or the size of the filter must
        be specified, but not both.
    :param size: Size of the filter, in bytes.
    :return: Dictionary to be inserted into the index.
    """

    if not (probability is None) ^ (size is None):
        raise ValueError("One of probability or size must be specified, but not both!")

    # We need k hash functions to generate a signature for every item, which we can construct
    # by taking a linear combination of two hash functions. The first hash function is a simple sha of the
    # column, the second one is a sha of the column + a deterministic salt.
    # First, we outsource the actual hashing to Postgres.

    # NULLs are interesting since a sha of a NULL is a NULL again. Here, we
    # turn them into strings with "NULL". This does mean they will collide if
    # we are indexing a string column with an actual "NULL" in it, but in this case
    # it will only mean chunks with NULLs will be fetched for a query with "NULL"
    # and vice versa, which doesn't break anything (this is just a preflight optimisation).

    digest_query = SQL(
        "SELECT digest(coalesce({0}::text, 'NULL'), 'sha256'), "
        "digest(coalesce({0}::text, 'NULL') || 'salt', 'sha256') "
        "FROM {1}.{2} o WHERE o.{3} = true"
    ).format(
        Identifier(column),
        Identifier(SPLITGRAPH_META_SCHEMA),
        Identifier(object_id),
        Identifier(SG_UD_FLAG),
    )

    digests = engine.run_sql(digest_query)

    # Add digests of the old values in the changeset for this column.
    if changeset:
        for _, old_row, _ in changeset.values():
            if column in old_row:
                # Note that this will add this value to the digest
                # even if it didn't get changed (say the row got overwritten
                # but another column got changed). This means that the value
                # in this case will be counted twice (both the new value
                # that we hashed before and the old value which we're
                # hashing now which is the same). This is not an issue
                # since we deduplicate our digests and the same digest
                # will set the same bits in the filter to 1, but something
                # to keep in mind.
                digests.append(_hash_value(old_row[column]))

    # Count the number of distinct items and determine the size (if needed) and optimal number
    # of hash functions.
    distinct_items = list(set(digests))

    if probability:
        # The formula gives the number of bits in the array, but we divide it by
        # 8 since we'll be using a byte array for operations + to store the signature.
        size = int(ceil(-len(distinct_items) * log(probability) / log(2) ** 2 / 8))

    # For mypy: we've asserted previously that either probability or size are specified
    # and we calculate the size from the probability, so size is no longer Optional[int].
    size = cast(int, size)

    size_bits = size * 8
    no_funcs = int(ceil(log(2) * size_bits / len(distinct_items)))

    # Generate the filter
    result = bytearray(size)
    for hash_1, hash_2 in distinct_items:
        hash_1 = int.from_bytes(hash_1, byteorder="big")
        hash_2 = int.from_bytes(hash_2, byteorder="big")
        for i in range(no_funcs):
            hash_i = (hash_1 + i * hash_2) % size_bits
            result[hash_i // 8] |= 1 << hash_i % 8

    return no_funcs, base64.b64encode(result).decode("ascii")


def describe(index_tuple: Tuple[int, str]) -> str:
    """
    Returns a pretty-printed summary of the bloom filter

    :param index_tuple: Tuple of (k, base64-encoded fingerprint)
        returned by generate_bloom_index
    :return: String
    """

    k, bloom_filter = index_tuple
    footprint = len(bloom_filter)
    bloom_filter = base64.b64decode(bloom_filter)

    # Get filter size in bytes (this is not the in-db footprint of the base64 value)
    filter_size = len(bloom_filter)

    # Calculate the number of set bits (used to approximate number of items)
    set_bits = 0
    for n in struct.unpack("B" * (len(bloom_filter)), bloom_filter):
        set_bits += bin(n).count("1")

    approx_items = -(filter_size * 8) / k * log(1 - set_bits / filter_size / 8)

    # Calculate the rough false positive probability
    probability = (1 - exp(-k * approx_items / filter_size / 8)) ** k

    return "k=%d, size %s, approx. %d item(s), false positive probability %.1f%%" % (
        k,
        pretty_size(footprint),
        int(approx_items),
        probability * 100,
    )


def _prepare_bloom_quals(quals: Any) -> List[List[Tuple[str, int, int]]]:
    """
    Convert list of qualifiers in CNF (ANDed OR-clauses where each clause is "column, operator, value")
    to prepare it for querying the bloom filter:

    * Clauses where operator isn't equality are set to True (we can't make a judgement on anything
      but exact matches).
    * Clauses with equality are converted to (column, hash_1, hash_2) so that the sought value
      isn't re-hashed for every fragment
    * OR-clauses where one operator is True are set to True completely (e.g. if we query
      a = 5 OR b > 6, the bloom filter can't say with certainty that there are no rows
      with b > 6 in the fragment and so we have to inspect it)
    * The toplevel AND-clause has all True values removed.

    :param quals: Quals in CNF.
    :return: Transformed list of quals
    """

    def _process_qual(qual):
        column, operator, value = qual
        if operator != "=":
            return True

        hash_1, hash_2 = _hash_value(value)
        hash_1 = int.from_bytes(hash_1, byteorder="big")
        hash_2 = int.from_bytes(hash_2, byteorder="big")

        return column, hash_1, hash_2

    def _process_or(quals):
        result = []
        for qual in quals:
            qual = _process_qual(qual)
            if qual is True:
                # anything OR True is True
                return True
            result.append(qual)
        return result

    result = []
    for or_quals in quals:
        or_quals = _process_or(or_quals)
        if or_quals is not True:
            result.append(or_quals)

    return result


def _match(qual: Tuple[str, int, int], bloom_index: Dict[str, Tuple[int, bytes]]) -> bool:
    """
    Checks whether a processed qual (column, hash_1, hash_2) can match a fragment with
    a given index.

    :param qual:
    :param bloom_index:
    """

    column, hash_1, hash_2 = qual
    if column not in bloom_index:
        # No index info for this column -- might match
        return True

    no_funcs, bloom_filter = bloom_index[column]
    size_bits = len(bloom_filter) * 8
    for i in range(no_funcs):
        hash_i = (hash_1 + i * hash_2) % size_bits
        if not bloom_filter[hash_i // 8] & (1 << hash_i % 8):
            # If at least one position in the filter isn't filled,
            # this qualifier can't be met by anything in the fragment.
            return False

    return True


def filter_bloom_index(engine: "PsycopgEngine", object_ids: List[str], quals: Any) -> List[str]:
    """
    Runs a bloom filter on given qualifiers using the given objects' previously-generated
    fingerprints.

    :param engine: Object engine
    :param object_ids: Object IDs
    :param quals: List of qualifiers
    :return: List of object IDs that might match the qualifiers in `quals` (including
        IDs that don't have a bloom index).
    """
    if not object_ids:
        return object_ids

    quals = _prepare_bloom_quals(quals)
    # If we don't have any equalities in quals or quals collapse to something
    # that the bloom filter can't make a judgement about, do nothing.
    if not quals:
        return object_ids

    # Load the index: my SQLfu isn't strong enough to create a query that takes
    # care of varying values of K and varying signature sizes.
    bloom_index = engine.run_sql(
        SQL(  # nosec
            "SELECT object_id, index -> 'bloom' FROM {}.{} WHERE object_id IN ("
            + ",".join(itertools.repeat("%s", len(object_ids)))
            + ")"
        ).format(Identifier(SPLITGRAPH_META_SCHEMA), Identifier("objects")),
        object_ids,
    )

    bloom_index = {
        o: {col: (i[0], base64.b64decode(i[1])) for col, i in index.items()}
        for o, index in bloom_index
        if index
    }

    dropped = []

    for object_id in object_ids:
        if object_id not in bloom_index:
            continue

        and_result = True
        for or_quals in quals:
            or_result = False
            for or_qual in or_quals:
                if _match(or_qual, bloom_index[object_id]):
                    or_result = True
                    break
            if not or_result:
                # One of the subclauses discarded this fragment.
                and_result = False
                break

        if not and_result:
            dropped.append(object_id)

    return [o for o in object_ids if o not in dropped]



import io
import itertools
import logging
import sys
import traceback
from operator import itemgetter
from typing import BinaryIO, Optional, TextIO, Tuple

import target_postgres
from target_postgres import DbSync
from target_postgres.db_sync import column_type, flatten_key

from splitgraph.config import CONFIG
from splitgraph.core.image import Image
from splitgraph.core.repository import Repository
from splitgraph.core.types import TableColumn, TableSchema
from splitgraph.engine.postgres.engine import get_change_key
from splitgraph.exceptions import TableNotFoundError
from splitgraph.ingestion.common import merge_tables

from .common import _make_changeset, _migrate_schema, log_exception, rollback_at_end


def select_breadcrumb(stream_message, breadcrumb):
    for sub_meta in stream_message["metadata"]:
        if sub_meta["breadcrumb"] == breadcrumb:
            return sub_meta["metadata"]

    raise ValueError("Breadcrumb %s not found!" % breadcrumb)


def get_key_properties(stream_message):
    """Extract the PK from a stream message. Supports both legacy ("key_properties") and
    new ("metadata") Singer taps."""
    if "key_properties" in stream_message:
        return stream_message["key_properties"]

    return select_breadcrumb(stream_message, []).get("table-key-properties", [])


class DbSyncProxy(DbSync):
    def __init__(self, *args, **kwargs):
        # The structure here is that we edit an image and write / modify tables in it. This
        # is supposed to be called with an image already existing.
        self.image = kwargs.pop("image")
        self.staging_schema = kwargs.pop("staging_schema")

        super().__init__(*args, **kwargs)

        self.staging_table: Optional[Tuple[str, str]] = None

    def _sg_schema(self) -> TableSchema:
        stream_schema_message = self.stream_schema_message
        return _get_sg_schema(self.flatten_schema, get_key_properties(stream_schema_message))

    def create_indices(self, stream):
        pass

    @rollback_at_end
    def sync_table(self):
        # NB the overridden method never calls self.update_columns() to bring the
        # schema up to date: this is because it compares a quoted name of the
        # table to the unquoted names returned by self.get_tables().

        schema_spec = self._sg_schema()

        # See if the table exists
        try:
            table = self.image.get_table(get_table_name(self.stream_schema_message))
        except TableNotFoundError:
            self.staging_table = (
                self.staging_schema,
                "staging_" + get_table_name(self.stream_schema_message),
            )
            self.logger.info("Creating a staging table at %s.%s", *self.staging_table)
            self.image.repository.object_engine.create_table(
                schema=self.staging_table[0],
                table=self.staging_table[1],
                schema_spec=schema_spec,
                unlogged=True,
            )
            # Make an empty table (will replace later on when flushing the data)
            self.image.repository.objects.register_tables(
                self.image.repository,
                [
                    (
                        self.image.image_hash,
                        get_table_name(self.stream_schema_message),
                        schema_spec,
                        [],
                    )
                ],
            )
            self.image.repository.commit_engines()

            return

        if table.table_schema != schema_spec:
            # Materialize the table into a temporary location and update its schema
            self.staging_table = (
                self.staging_schema,
                "staging_" + get_table_name(self.stream_schema_message),
            )
            self.logger.info(
                "Schema mismatch, materializing the table into %s.%s and migrating",
                *self.staging_table,
            )
            table.materialize(self.staging_table[1], self.staging_table[0])

            _migrate_schema(
                self.image.repository.object_engine,
                self.staging_table[0],
                self.staging_table[1],
                table.table_schema,
                schema_spec,
            )
        self.image.repository.commit_engines()

    @log_exception
    @rollback_at_end
    def load_csv(self, file, count, size_bytes):
        from splitgraph.ingestion.csv import copy_csv_buffer

        table_name = get_table_name(self.stream_schema_message)
        schema_spec = self._sg_schema()
        temp_table = "tmp_" + table_name
        self.logger.info("Loading %d rows into '%s'", count, table_name)

        old_table = self.image.get_table(table_name)

        self.image.repository.object_engine.create_table(
            schema="pg_temp", table=temp_table, schema_spec=schema_spec, temporary=True
        )

        with open(file, "rb") as f:
            copy_csv_buffer(
                data=f,
                engine=self.image.repository.object_engine,
                schema="pg_temp",
                table=temp_table,
                no_header=True,
                escape="\\",
            )
        schema_spec = self._sg_schema()
        if not self.staging_table:
            self._merge_existing_table(old_table, temp_table)
        else:
            self._overwrite_existing_table(temp_table, table_name, schema_spec)

        # Commit (deletes temporary tables)
        self.image.repository.commit_engines()

    def _overwrite_existing_table(self, temp_table, table_name, schema_spec):
        assert self.staging_table
        staging_table_schema, staging_table = self.staging_table

        # Merge the CSV data into the "materialized" table
        merge_tables(
            self.image.object_engine,
            "pg_temp",
            temp_table,
            schema_spec,
            staging_table_schema,
            staging_table,
            schema_spec,
        )
        object_id = self.image.repository.objects.create_base_fragment(
            staging_table_schema,
            staging_table,
            self.image.repository.namespace,
            table_schema=schema_spec,
        )

        # Overwrite the existing table
        self.image.repository.objects.overwrite_table(
            self.image.repository, self.image.image_hash, table_name, schema_spec, [object_id]
        )
        self.staging_table = None

    def _merge_existing_table(self, old_table, temp_table):
        # Load the data directly as a Splitgraph object. Note that this can still be
        # improved: currently the Singer loader loads the stream as a CSV, gives it
        # to us, we ingest it into a temporary table, generate a changeset from that
        # (which goes back into Python), give it to the fragment manager, the fragment
        # manager builds an object by loading it into a temporary table _again_ and then
        # into a cstore_fdw file.

        assert self._sg_schema() == old_table.table_schema
        with old_table.image.query_schema(commit=False):
            # If hard_delete is set, check the sdc_removed_at flag in the table: if it's
            # not NULL, we mark rows as deleted instead.
            hard_delete = bool(self.connection_config.get("hard_delete"))

            # Find PKs that have been upserted and deleted (make fake changeset)
            changeset = _make_changeset(
                self.image.object_engine,
                "pg_temp",
                temp_table,
                old_table.table_schema,
                upsert_condition="_sdc_removed_at IS NOT DISTINCT FROM NULL"
                if hard_delete
                else "TRUE",
            )

            inserted = sum(1 for v in changeset.values() if v[0] and not v[1])
            updated = sum(1 for v in changeset.values() if v[0] and v[1])
            deleted = sum(1 for v in changeset.values() if not v[0])
            self.logger.info(
                "Table %s: inserted %d, updated %d, deleted %d",
                get_table_name(self.stream_schema_message),
                inserted,
                updated,
                deleted,
            )

            # Split the changeset according to the original fragment boundaries so that a single
            # change spanning multiple fragments doesn't force materialization.
            split_changesets = self.image.repository.objects.split_changeset_boundaries(
                changeset, get_change_key(old_table.table_schema), old_table.objects
            )
            self.logger.info(
                "Table %s: split changeset into %d fragment(s)",
                get_table_name(self.stream_schema_message),
                len(split_changesets),
            )

            # Store the changeset as a new SG object
            object_ids = self.image.repository.objects._store_changesets(
                old_table, split_changesets, "pg_temp", table_name=temp_table
            )

        # Add the new object to the table.
        # add_table (called by register_tables) does that by default (appends
        # the object to the table if it already exists)
        self.image.repository.objects.register_tables(
            self.image.repository,
            [
                (
                    self.image.image_hash,
                    old_table.table_name,
                    old_table.table_schema,
                    object_ids,
                )
            ],
        )

    def delete_rows(self, stream):
        # We delete rows in load_csv if required (by putting them into the DIFF as
        # upserted=False).
        pass

    def create_schema_if_not_exists(self, table_columns_cache=None):
        pass


def _get_sg_schema(flattened_schema, primary_key) -> TableSchema:
    return [
        TableColumn(i, name, column_type(schema_property), name in primary_key, None)
        for i, (name, schema_property) in enumerate(flattened_schema.items())
    ]


# Taken from target-postgres and adapted to not crash on unsupported columns
def _flatten_schema(d, parent_key=None, sep="__", level=0, max_level=0):
    if parent_key is None:
        parent_key = []

    items = []

    if "properties" not in d:
        return {}

    for k, v in d["properties"].items():
        new_key = flatten_key(k, parent_key, sep)
        if "type" in v:
            if "object" in v["type"] and "properties" in v and level < max_level:
                items.extend(
                    _flatten_schema(
                        v, parent_key + [k], sep=sep, level=level + 1, max_level=max_level
                    ).items()
                )
            else:
                items.append((new_key, v))
        else:
            if len(v.values()) > 0:
                if v.get("inclusion") == "unsupported":
                    logging.warning("Unsupported field %s: %s", new_key, v.get("description", ""))
                    continue
                if list(v.values())[0][0]["type"] == "string":
                    list(v.values())[0][0]["type"] = ["null", "string"]
                    items.append((new_key, list(v.values())[0][0]))
                elif list(v.values())[0][0]["type"] == "array":
                    list(v.values())[0][0]["type"] = ["null", "array"]
                    items.append((new_key, list(v.values())[0][0]))
                elif list(v.values())[0][0]["type"] == "object":
                    list(v.values())[0][0]["type"] = ["null", "object"]
                    items.append((new_key, list(v.values())[0][0]))

    sorted_items = sorted(items, key=itemgetter(0))
    for k, g in itertools.groupby(sorted_items, key=itemgetter(0)):
        if len(list(g)) > 1:
            raise ValueError("Duplicate column name produced in schema: {}".format(k))

    return dict(sorted_items)


def get_sg_schema(stream_schema_message, flattening_max_level=0):
    return _get_sg_schema(
        _flatten_schema(stream_schema_message["schema"], max_level=flattening_max_level),
        get_key_properties(stream_schema_message),
    )


def get_table_name(stream_schema_message):
    return stream_schema_message["stream"].replace(".", "_").replace("-", "_").lower()


def db_sync_wrapper(image: "Image", staging_schema: str):
    def wrapped(*args, **kwargs):
        return DbSyncProxy(image=image, staging_schema=staging_schema, *args, **kwargs)

    return wrapped


def run_patched_sync(
    repository: Repository,
    base_image: Optional[Image],
    new_image_hash: str,
    delete_old: bool,
    failure: str,
    input_stream: Optional[BinaryIO] = None,
    output_stream: Optional[TextIO] = None,
):
    input_stream = input_stream or sys.stdin.buffer

    # Build a staging schema
    staging_schema = "sg_tmp_" + repository.to_schema()
    repository.object_engine.delete_schema(staging_schema)
    repository.object_engine.create_schema(staging_schema)
    repository.commit_engines()

    config = _prepare_config_params(repository)
    old_sync = target_postgres.DbSync

    stdout = sys.stdout
    target_postgres.DbSync = db_sync_wrapper(repository.images[new_image_hash], staging_schema)
    if output_stream:
        sys.stdout = output_stream
    try:
        singer_messages = io.TextIOWrapper(input_stream, encoding="utf-8")
        target_postgres.persist_lines(config, singer_messages)
        if delete_old and base_image:
            repository.images.delete([base_image.image_hash])
    except Exception:
        repository.rollback_engines()
        if failure == "delete-new":
            repository.images.delete([new_image_hash])
        elif failure == "delete-old" and base_image:
            repository.images.delete([base_image.image_hash])
        repository.commit_engines()
        logging.error(traceback.format_exc())
        raise
    finally:
        sys.stdout = stdout
        target_postgres.DbSync = old_sync
        repository.object_engine.delete_schema(staging_schema)
        repository.commit_engines()


def _prepare_config_params(repository):
    conn_params = repository.engine.conn_params
    # Prepare target_postgres config
    config = {
        "host": conn_params["SG_ENGINE_HOST"],
        "port": int(conn_params["SG_ENGINE_PORT"]),
        "user": conn_params["SG_ENGINE_USER"],
        "password": conn_params["SG_ENGINE_PWD"],
        "dbname": conn_params["SG_ENGINE_DB_NAME"],
        "default_target_schema": repository.to_schema(),
        "max_parallelism": _calc_max_threads(conn_params),
    }
    return config


def _calc_max_threads(conn_params):
    """Each loader thread really uses 2 connections (one directly when communicating with
    the engine, one from the in-engine Splitgraph during querying) and this can lead to
    weird deadlocks when we exhaust the connection limit enforced by pgbouncer) -- to avoid
    that, we decrease the number of concurrent loader threads."""
    return max(int(conn_params.get("SG_ENGINE_POOL", CONFIG["SG_ENGINE_POOL"])) // 2 - 1, 1)



"""
Definitions for responses from the cloud GQL/REST APIs
"""
import enum
import logging
from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from splitgraph.cloud.project.models import (
    External,
    IngestionSchedule,
    Metadata,
    Repository,
    Source,
    Table,
)
from splitgraph.core.types import MountError, Params, TableSchema


class Plugin(BaseModel):
    plugin_name: str
    credentials_schema: Dict[str, Any]
    params_schema: Dict[str, Any]
    table_params_schema: Dict[str, Any]
    name: str
    description: str
    supports_mount: bool
    supports_load: bool
    supports_sync: bool


class IntrospectionMode(str, enum.Enum):
    """
    Which tables to (re)introspect when adding an external.
    """

    NONE = "none"  # Don't reintrospect any tables
    EMPTY = "empty"  # Introspect tables with an empty schema
    ALL = "all"  # Reintrospect all tables


# GQL response for the catalog metadata


class MetadataResponse(BaseModel):
    class RepoTopicsResponse(BaseModel):
        class RepositoryTopics(BaseModel):
            topic: str

        nodes: List[RepositoryTopics]

    class RepoProfileResponse(BaseModel):
        class MetadataResponse(BaseModel):
            created_at: Optional[str]
            updated_at: Optional[str]
            upstream_metadata: Dict[str, Any]

        readme: Optional[str]
        description: Optional[str]
        topics: Optional[List[str]]
        sources: Optional[List[Source]]
        license: Optional[str]
        metadata: Optional[MetadataResponse]

    namespace: str
    repository: str
    repoTopicsByNamespaceAndRepository: RepoTopicsResponse
    repoProfileByNamespaceAndRepository: RepoProfileResponse

    @classmethod
    def from_response(cls, response) -> List["MetadataResponse"]:
        nodes = response["data"]["repositories"]["nodes"]
        return [cls.parse_obj(obj) for obj in nodes]

    def to_metadata(self) -> Metadata:
        profile = self.repoProfileByNamespaceAndRepository

        return Metadata(
            readme=Metadata.Readme(text=profile.readme),
            description=profile.description,
            topics=[node.topic for node in self.repoTopicsByNamespaceAndRepository.nodes],
            sources=profile.sources,
            license=profile.license,
            extra_metadata=profile.metadata.upstream_metadata if profile.metadata else None,
        )


# GQL response for the external repo's metadata


class ExternalResponse(BaseModel):
    class ExternalImageResponse(BaseModel):
        class ImageResponse(BaseModel):
            class TablesResponse(BaseModel):
                class Node(BaseModel):
                    tableName: str
                    tableSchema: TableSchema

                nodes: List[Node]

            tablesByNamespaceAndRepositoryAndImageHash: TablesResponse

        imageByNamespaceAndRepositoryAndImageHash: ImageResponse

    class IngestionScheduleResponse(BaseModel):
        schedule: str
        enabled = True
        schema_: Dict[str, TableSchema] = Field(alias="schema")

    namespace: str
    repository: str
    credentialId: Optional[str]
    dataSource: str
    params: Dict[str, Any]
    tableParams: Dict[str, Any]
    externalImageByNamespaceAndRepository: Optional[ExternalImageResponse]
    ingestionScheduleByNamespaceAndRepository: Optional[IngestionScheduleResponse]

    @classmethod
    def from_response(cls, response) -> List["ExternalResponse"]:
        nodes = response["data"]["repositoryDataSources"]["nodes"]
        return [cls.parse_obj(obj) for obj in nodes]

    def to_external(self) -> External:
        schemas: Dict[str, TableSchema] = {}
        ingestion_schedule: Optional[IngestionSchedule] = None

        if self.ingestionScheduleByNamespaceAndRepository:
            schemas = {
                tn: ts for tn, ts in self.ingestionScheduleByNamespaceAndRepository.schema_.items()
            }
            ingestion_schedule = IngestionSchedule(
                schedule=self.ingestionScheduleByNamespaceAndRepository.schedule,
                enabled=self.ingestionScheduleByNamespaceAndRepository.enabled,
            )

        if self.externalImageByNamespaceAndRepository:
            schemas = {
                t.tableName: t.tableSchema
                for t in self.externalImageByNamespaceAndRepository.imageByNamespaceAndRepositoryAndImageHash.tablesByNamespaceAndRepositoryAndImageHash.nodes
            }

        if not schemas:
            logging.warning(
                "No table schemas found for repository %s/%s", self.namespace, self.repository
            )

        all_tables = list(self.tableParams.keys())
        all_tables.extend(schemas.keys())

        tables = {
            t: Table(
                options=self.tableParams.get(t, {}),
                schema=[Table.Column(name=c.name, type=c.pg_type) for c in schemas.get(t, [])],
            )
            for t in all_tables
        }

        return External(
            credential_id=self.credentialId,
            plugin=self.dataSource,
            params=self.params,
            tables=tables,
            schedule=ingestion_schedule,
        )


class JobStatus(BaseModel):
    task_id: str
    started: datetime
    finished: Optional[datetime]
    status: Optional[str]


class IngestionJobStatus(JobStatus):
    is_manual: bool


class ExportJobStatus(JobStatus):
    user_id: Optional[str]
    export_format: str
    output: Optional[Dict[str, Any]]


class RepositoryIngestionJobStatusResponse(BaseModel):
    class RepositoryIngestionJobStatus(BaseModel):
        class Node(BaseModel):
            taskId: str
            started: datetime
            finished: Optional[datetime]
            isManual: bool
            status: Optional[str]

        nodes: List[Node]

    repositoryIngestionJobStatus: RepositoryIngestionJobStatus

    @classmethod
    def from_response(cls, response) -> "RepositoryIngestionJobStatusResponse":
        return cls.parse_obj(response["data"])


def make_repositories(
    metadata_responses: List[MetadataResponse], external_responses: List[ExternalResponse]
) -> List[Repository]:
    metadata = {(m.namespace, m.repository): m.to_metadata() for m in metadata_responses}
    external = {(r.namespace, r.repository): r.to_external() for r in external_responses}

    return [
        Repository(
            namespace=n, repository=r, metadata=metadata.get((n, r)), external=external.get((n, r))
        )
        for (n, r) in sorted(set(metadata.keys()) | set(external.keys()))
    ]


class ListExternalCredentialsResponse(BaseModel):
    class ExternalCredential(BaseModel):
        plugin_name: str
        credential_name: Optional[str]
        credential_id: str

    credentials: List[ExternalCredential]


class UpdateExternalCredentialRequest(BaseModel):
    credential_id: str
    credential_name: str
    credential_data: Dict[str, Any]
    plugin_name: str


class AddExternalCredentialRequest(BaseModel):
    credential_name: str
    credential_data: Dict[str, Any]
    plugin_name: str


class UpdateExternalCredentialResponse(BaseModel):
    credential_id: str


class ExternalTableRequest(BaseModel):
    options: Dict[str, Any] = {}
    schema_: Optional[Dict[str, str]] = Field(alias="schema")


class AddExternalRepositoryRequest(BaseModel):
    namespace: str
    repository: str
    plugin_name: str
    params: Params
    is_live: bool
    tables: Optional[Dict[str, ExternalTableRequest]]
    credential_id: Optional[str]
    credential_name: Optional[str]
    credential_data: Optional[Dict[str, Any]]
    schedule: Optional[IngestionSchedule]
    initial_private: bool = False

    @classmethod
    def from_external(
        cls,
        namespace: str,
        repository: str,
        external: External,
        credential_map: Optional[Dict[str, str]] = None,
        initial_private: bool = False,
    ):
        credential_map = credential_map or {}

        credential_id = external.credential_id
        if external.credential and not credential_id:
            if external.credential not in credential_map:
                raise ValueError(
                    "Credential %s not defined in the file and not set up on the remote!"
                    % external.credential
                )
            credential_id = credential_map[external.credential]

        return cls(
            namespace=namespace,
            repository=repository,
            plugin_name=external.plugin,
            params=external.params,
            tables={
                table_name: ExternalTableRequest(
                    options=table.options, schema={c.name: c.pg_type for c in table.schema_}
                )
                for table_name, table in external.tables.items()
            },
            credential_id=credential_id,
            is_live=external.is_live,
            schedule=external.schedule,
            initial_private=initial_private,
        )


class AddExternalRepositoriesRequest(BaseModel):
    repositories: List[AddExternalRepositoryRequest]
    introspection_mode: IntrospectionMode = IntrospectionMode.EMPTY


class AddExternalRepositoriesResponse(BaseModel):
    class RepositoryMountError(BaseModel):
        namespace: str
        repository: str
        errors: List[MountError]

    live_image_hashes: List[Optional[str]]
    errors: Optional[List[RepositoryMountError]] = None



# List of permitted AST toplevel statements. In IMPORTs (e.g. sgr import
# and Splitfile IMPORT commands), we only allow read-only SELECTs. In Splitfile
# SQL commands, we also allow users to use a bunch of writeable DML and DDL.

IMPORT_SQL_PERMITTED_STATEMENTS = ["RawStmt", "SelectStmt"]
SPLITFILE_SQL_PERMITTED_STATEMENTS = IMPORT_SQL_PERMITTED_STATEMENTS + [
    "InsertStmt",
    "UpdateStmt",
    "DeleteStmt",
    "CreateStmt",
    "CreateTableAsStmt",
    "AlterTableStmt",
    "DropStmt",
]


# List of table names/views/indexes that are located in pg_catalog. This is always
# available to SQL statements even if the search_path doesn't have pg_catalog in it.
# Generating this automatically requires a connection to the engine, adding a dependency
# on that to the sql module, so we hardcode this.

# Generated by
# sgr sql "SELECT relname FROM pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE n.nspname = 'pg_catalog'" -a > blacklist.txt

PG_CATALOG_TABLES = set(
    """pg_statistic
pg_type
pg_foreign_server
pg_aggregate_fnoid_index
pg_am_name_index
pg_am_oid_index
pg_amop_fam_strat_index
pg_amop_opr_fam_index
pg_amop_oid_index
pg_amproc_fam_proc_index
pg_amproc_oid_index
pg_attrdef_adrelid_adnum_index
pg_attrdef_oid_index
pg_attribute_relid_attnam_index
pg_attribute_relid_attnum_index
pg_authid_rolname_index
pg_authid_oid_index
pg_auth_members_role_member_index
pg_auth_members_member_role_index
pg_cast_oid_index
pg_cast_source_target_index
pg_class_oid_index
pg_class_relname_nsp_index
pg_class_tblspc_relfilenode_index
pg_collation_name_enc_nsp_index
pg_collation_oid_index
pg_constraint_conname_nsp_index
pg_constraint_conrelid_contypid_conname_index
pg_constraint_contypid_index
pg_constraint_oid_index
pg_constraint_conparentid_index
pg_conversion_default_index
pg_conversion_name_nsp_index
pg_conversion_oid_index
pg_database_datname_index
pg_database_oid_index
pg_depend_depender_index
pg_depend_reference_index
pg_description_o_c_o_index
pg_shdescription_o_c_index
pg_enum_oid_index
pg_enum_typid_label_index
pg_enum_typid_sortorder_index
pg_index_indrelid_index
pg_index_indexrelid_index
pg_inherits_relid_seqno_index
pg_inherits_parent_index
pg_init_privs_o_c_o_index
pg_language_name_index
pg_language_oid_index
pg_largeobject_loid_pn_index
pg_largeobject_metadata_oid_index
pg_namespace_nspname_index
pg_namespace_oid_index
pg_opclass_am_name_nsp_index
pg_opclass_oid_index
pg_operator_oid_index
pg_operator_oprname_l_r_n_index
pg_opfamily_am_name_nsp_index
pg_opfamily_oid_index
pg_pltemplate_name_index
pg_proc_oid_index
pg_proc_proname_args_nsp_index
pg_rewrite_oid_index
pg_rewrite_rel_rulename_index
pg_sequence_seqrelid_index
pg_shdepend_depender_index
pg_shdepend_reference_index
pg_statistic_relid_att_inh_index
pg_statistic_ext_oid_index
pg_statistic_ext_name_index
pg_statistic_ext_relid_index
pg_statistic_ext_data_stxoid_index
pg_tablespace_oid_index
pg_tablespace_spcname_index
pg_transform_oid_index
pg_transform_type_lang_index
pg_trigger_tgconstraint_index
pg_trigger_tgrelid_tgname_index
pg_trigger_oid_index
pg_event_trigger_evtname_index
pg_event_trigger_oid_index
pg_ts_config_cfgname_index
pg_ts_config_oid_index
pg_ts_config_map_index
pg_ts_dict_dictname_index
pg_ts_dict_oid_index
pg_ts_parser_prsname_index
pg_ts_parser_oid_index
pg_ts_template_tmplname_index
pg_ts_template_oid_index
pg_type_oid_index
pg_type_typname_nsp_index
pg_foreign_data_wrapper_oid_index
pg_foreign_data_wrapper_name_index
pg_foreign_server_oid_index
pg_foreign_server_name_index
pg_user_mapping_oid_index
pg_user_mapping_user_server_index
pg_foreign_table_relid_index
pg_default_acl_role_nsp_obj_index
pg_default_acl_oid_index
pg_db_role_setting_databaseid_rol_index
pg_seclabel_object_index
pg_shseclabel_object_index
pg_extension_oid_index
pg_extension_name_index
pg_range_rngtypid_index
pg_policy_oid_index
pg_policy_polrelid_polname_index
pg_replication_origin_roiident_index
pg_replication_origin_roname_index
pg_partitioned_table_partrelid_index
pg_publication_oid_index
pg_publication_pubname_index
pg_publication_rel_oid_index
pg_publication_rel_prrelid_prpubid_index
pg_subscription_oid_index
pg_subscription_subname_index
pg_subscription_rel_srrelid_srsubid_index
pg_authid
pg_shadow
pg_statistic_ext_data
pg_roles
pg_settings
pg_file_settings
pg_hba_file_rules
pg_config
pg_cursors
pg_user_mapping
pg_stat_bgwriter
pg_replication_origin_status
pg_subscription
pg_stat_progress_vacuum
pg_stat_progress_cluster
pg_attribute
pg_proc
pg_class
pg_attrdef
pg_constraint
pg_inherits
pg_index
pg_operator
pg_opfamily
pg_opclass
pg_am
pg_amop
pg_amproc
pg_language
pg_largeobject_metadata
pg_aggregate
pg_stat_progress_create_index
pg_user_mappings
pg_largeobject
pg_statistic_ext
pg_rewrite
pg_trigger
pg_event_trigger
pg_description
pg_cast
pg_enum
pg_namespace
pg_conversion
pg_depend
pg_database
pg_db_role_setting
pg_tablespace
pg_pltemplate
pg_auth_members
pg_shdepend
pg_shdescription
pg_ts_config
pg_ts_config_map
pg_ts_dict
pg_ts_parser
pg_ts_template
pg_extension
pg_foreign_data_wrapper
pg_foreign_table
pg_policy
pg_replication_origin
pg_default_acl
pg_init_privs
pg_seclabel
pg_shseclabel
pg_collation
pg_partitioned_table
pg_range
pg_transform
pg_sequence
pg_publication
pg_publication_rel
pg_subscription_rel
pg_group
pg_user
pg_policies
pg_rules
pg_views
pg_tables
pg_matviews
pg_indexes
pg_sequences
pg_stats
pg_stats_ext
pg_publication_tables
pg_locks
pg_available_extensions
pg_available_extension_versions
pg_prepared_xacts
pg_prepared_statements
pg_seclabels
pg_statio_sys_tables
pg_timezone_abbrevs
pg_timezone_names
pg_statio_user_tables
pg_stat_all_tables
pg_stat_xact_all_tables
pg_stat_sys_tables
pg_stat_xact_sys_tables
pg_stat_user_tables
pg_stat_xact_user_tables
pg_statio_all_tables
pg_stat_all_indexes
pg_stat_sys_indexes
pg_stat_user_indexes
pg_statio_all_indexes
pg_statio_sys_indexes
pg_statio_user_indexes
pg_statio_all_sequences
pg_statio_sys_sequences
pg_statio_user_sequences
pg_stat_activity
pg_stat_replication
pg_stat_wal_receiver
pg_stat_subscription
pg_stat_ssl
pg_stat_gssapi
pg_replication_slots
pg_stat_database
pg_stat_database_conflicts
pg_stat_user_functions
pg_stat_xact_user_functions
pg_stat_archiver
""".split()
)



"""
sgr commands related to creating and checking out images
"""
import sys
from collections import defaultdict

import click

from splitgraph.commandline.common import (
    ImageType,
    JsonType,
    RepositoryType,
    remote_switch_option,
)
from splitgraph.config import DEFAULT_CHUNK_SIZE
from splitgraph.exceptions import TableNotFoundError


@click.command(name="checkout")
@click.argument("image_spec", type=ImageType(default="HEAD", get_image=True))
@click.option(
    "-f", "--force", help="Discard all pending changes to the schema", is_flag=True, default=False
)
@click.option(
    "-u", "--uncheckout", help="Delete the checked out copy instead", is_flag=True, default=False
)
@click.option(
    "-l",
    "--layered",
    help="Don't materialize the tables, use layered querying instead.",
    is_flag=True,
    default=False,
)
def checkout_c(image_spec, force, uncheckout, layered):
    """
    Check out a Splitgraph image into a Postgres schema.

    This downloads the required physical objects and materializes all tables, unless ``-l`` or ``--layered`` is passed,
    in which case the objects are downloaded and a foreign data wrapper is set up on the engine to satisfy read-only
    queries by combining results from each table's fragments.

    Tables checked out in this way are still presented as normal Postgres tables and can queried in the same way.
    Since the tables aren't materialized, layered querying is faster to set up, but since each query now results in a
    subquery to each object comprising the table, actual query execution is slower than to materialized Postgres tables.

    Layered querying is only supported for read-only queries.

    Image spec must be of the format ``[NAMESPACE/]REPOSITORY[:HASH_OR_TAG]``. Note that currently, the schema that the
    image is checked out into has to have the same name as the repository. If no image hash or tag is passed,
    "HEAD" is assumed.

    If ``-u`` or ``--uncheckout`` is passed, this instead deletes the checked out schema (assuming there are no pending
    changes) and removes the HEAD pointer.

    If ``--force`` isn't passed and the schema has pending changes, this will fail.
    """
    repository, image = image_spec

    if uncheckout:
        repository.uncheckout(force=force)
        click.echo("Unchecked out %s." % (str(repository),))
    else:
        image.checkout(force=force, layered=layered)
        click.echo("Checked out %s:%s." % (str(repository), image.image_hash[:12]))


@click.command(name="commit")
@click.argument("repository", type=RepositoryType(exists=True))
@click.option(
    "-s",
    "--snap",
    default=False,
    is_flag=True,
    help="Do not delta compress the changes and instead store the whole table again. "
    "This consumes more space, but makes checkouts faster.",
)
@click.option(
    "-c",
    "--chunk-size",
    default=DEFAULT_CHUNK_SIZE,
    type=int,
    help="Split new tables into chunks of this many rows (by primary key). The default "
    "value is governed by the SG_COMMIT_CHUNK_SIZE configuration parameter.",
)
@click.option(
    "-k",
    "--chunk-sort-keys",
    default=None,
    type=JsonType(),
    help="Sort the data inside each chunk by this/these key(s)",
)
@click.option(
    "-t",
    "--split-changesets",
    default=False,
    is_flag=True,
    help="Split changesets for existing tables across original chunk boundaries.",
)
@click.option(
    "-i",
    "--index-options",
    type=JsonType(),
    help="JSON dictionary of extra indexes to calculate on the new objects.",
)
@click.option("-m", "--message", help="Optional commit message")
@click.option(
    "-o", "--overwrite", is_flag=True, help="Overwrite physical objects that already exist"
)
def commit_c(
    repository,
    snap,
    chunk_size,
    chunk_sort_keys,
    split_changesets,
    index_options,
    message,
    overwrite,
):
    """
    Commit changes to a checked-out Splitgraph repository.

    This packages up all changes into a new image. Where a table hasn't been created or had its schema changed,
    this will delta compress the changes. For all other tables (or if ``-s`` has been passed), this will
    store them as full table snapshots.

    When a table is stored as a full snapshot, `--chunk-size` sets the maximum size, in rows, of the fragments
    that the table will be split into (default is no splitting). The splitting is done by the
    table's primary key.

    If `--split-changesets` is passed, delta-compressed changes will also be split up according to the original
    table chunk boundaries. For example, if there's a change to the first and the 20000th row of a table that was
    originally committed with `--chunk-size=10000`, this will create 2 fragments: one based on the first chunk
    and one on the second chunk of the table.

    If `--chunk-sort-keys` is passed, data inside the chunk is sorted by this key (or multiple keys).
    This helps speed up queries on those keys for storage layers than can leverage that (e.g. CStore). The expected format is JSON, e.g. `{table_1: [col_1, col_2]}`

    `--index-options` expects a JSON-serialized dictionary of `{table: index_type: column: index_specific_kwargs}`.
    Indexes are used to narrow down the amount of chunks to scan through when running a query. By default, each column
    has a range index (minimum and maximum values) and it's possible to add bloom filtering to speed up queries that
    involve equalities.

    Bloom filtering allows to trade off between the space overhead of the index and the probability of a false
    positive (claiming that an object contains a record when it actually doesn't, leading to extra scans).

    An example `index-options` dictionary:

    \b
    ```
    {
        "table": {
            "bloom": {
                "column_1": {
                    "probability": 0.01,   # Only one of probability
                    "size": 10000          # or size can be specified.
                }
            },
            # Only compute the range index on these columns. By default,
            # it's computed on all columns and is always computed on the
            # primary key no matter what.
            "range": ["column_2", "column_3"]
        }
    }
    ```
    """
    new_hash = repository.commit(
        comment=message,
        snap_only=snap,
        chunk_size=chunk_size,
        split_changeset=split_changesets,
        extra_indexes=index_options,
        in_fragment_order=chunk_sort_keys,
        overwrite=overwrite,
    ).image_hash
    click.echo("Committed %s as %s." % (str(repository), new_hash[:12]))


@click.command(name="tag")
@click.argument("image_spec", type=ImageType(default=None))
@click.argument("tag", required=False)
@click.option("-d", "--delete", is_flag=True, help="Delete the tag instead.")
@remote_switch_option()
def tag_c(image_spec, tag, delete):
    """
    Manage tags on images.

    Depending on the exact invocation, this command can tag a Splitgraph image,
    list all tags in a repository or delete a tag.

    Examples:

    ``sgr tag noaa/climate``

    List all tagged images in the ``noaa/climate`` repository and their tags.

    ``sgr tag noaa/climate:abcdef1234567890``

    List all tags assigned to the image ``noaa/climate:abcdef1234567890...``

    ``sgr tag noaa/climate:abcdef1234567890 my_new_tag``

    Tag the image ``noaa/climate:abcdef1234567890...`` with ``my_new_tag``. If the tag already exists, this will
    overwrite the tag.

    ``sgr tag noaa/climate my_new_tag``

    Tag the current ``HEAD`` of ``noaa/climate`` with ``my_new_tag``.

    ``sgr tag --delete noaa/climate:my_new_tag``

    Delete the tag ``my_new_tag`` from ``noaa/climate``.
    """
    repository, image = image_spec

    if delete:
        # In this case the tag must be a part of the image spec.
        if tag is not None or image is None:
            raise click.BadArgumentUsage(
                "Use sgr tag --delete %s:TAG_TO_DELETE" % repository.to_schema()
            )
        if image in ("latest", "HEAD"):
            raise click.BadArgumentUsage("%s is a reserved tag!" % image)
        repository.images[image].delete_tag(image)
        return

    if tag is None:
        # List all tags
        tag_dict = defaultdict(list)
        for img, img_tag in repository.get_all_hashes_tags():
            tag_dict[img].append(img_tag)
        if image is None:
            for img, tags in tag_dict.items():
                # Sometimes HEAD is none (if we've just cloned the repo)
                if img:
                    click.echo("%s: %s" % (img[:12], ", ".join(sorted(tags))))
        else:
            click.echo(", ".join(tag_dict[repository.images[image].image_hash]))
        return

    if tag == "HEAD":
        raise click.BadArgumentUsage("HEAD is a reserved tag!")

    if image is None:
        image = repository.head
    else:
        image = repository.images[image]

    image.tag(tag)
    click.echo("Tagged %s:%s with %s." % (str(repository), image.image_hash, tag))


@click.command(name="import")
@click.argument("image_spec", type=ImageType())
@click.argument("table_or_query")
@click.argument("target_repository", type=RepositoryType())
@click.argument("target_table", required=False)
def import_c(image_spec, table_or_query, target_repository, target_table):
    """
    Import tables into a Splitgraph repository.

    Imports a table or a result of a query from a local Splitgraph repository or a Postgres schema into another
    Splitgraph repository.

    Examples:

    ``sgr import noaa/climate:my_tag climate_data my/repository``

    Create a new image in ``my/repository`` with the ``climate_data`` table included. This links the new image to
    the physical object, meaning that the history of the ``climate_data`` table is preserved.

    If no tag is specified, the 'latest' (not the HEAD image or current state of the checked out image)
    image is used.

    ``sgr import noaa/climate:my_tag "SELECT * FROM climate_data" my/repository climate_data``

    Create a new image in ``my/repository`` with the result of the query stored in the ``climate_data`` table. This
    creates a new physical object without any linkage to the original data, so the history of the ``climate_data``
    table isn't preserved. The SQL query can interact with multiple tables in the source image.

    ``sgr import other_schema other_table my/repository``

    Since other_schema isn't a Splitgraph repository, this will copy ``other_schema.other_table``
    into a new Splitgraph object and add the ``other_table`` table to a new image in ``my/repository``.

    Note that importing doesn't discard or commit pending changes in the target Splitgraph repository: a new image
    is created with the new table added, the new table is materialized in the repository and the HEAD pointer is moved.
    """
    from splitgraph.core.engine import repository_exists

    repository, image = image_spec

    if repository_exists(repository):
        foreign_table = False
        image = repository.images[image]
        # If the source table doesn't exist in the image, we'll treat it as a query instead.
        try:
            image.get_table(table_or_query)
            is_query = False
        except TableNotFoundError:
            is_query = True
    else:
        # If the source schema isn't actually a Splitgraph repo, we'll be copying the table verbatim.
        foreign_table = True
        is_query = table_or_query not in repository.engine.get_all_tables(repository.to_schema())
        image = None

    if is_query and not target_table:
        click.echo("TARGET_TABLE is required when the source is a query!")
        sys.exit(1)

    target_repository.import_tables(
        [target_table] if target_table else [],
        repository,
        [table_or_query],
        image_hash=image.image_hash if image else None,
        foreign_tables=foreign_table,
        table_queries=[] if not is_query else [True],
    )

    click.echo(
        "%s:%s has been imported from %s:%s%s"
        % (
            str(target_repository),
            target_table,
            str(repository),
            table_or_query,
            (" (%s)" % image.image_hash[:12] if image else ""),
        )
    )


@click.command(name="reindex")
@click.argument("image_spec", type=ImageType(default="HEAD", get_image=True))
@click.argument("table_name", type=str)
@click.option(
    "-i",
    "--index-options",
    type=JsonType(),
    required=True,
    help="JSON dictionary of extra indexes to calculate, e.g. "
    '\'{"bloom": {"column_1": {"probability": 0.01}}}\'',
)
@click.option(
    "-o",
    "--ignore-patch-objects",
    type=bool,
    is_flag=True,
    default=False,
    help="Ignore objects that change other objects' rows instead of raising an error",
)
def reindex_c(image_spec, table_name, index_options, ignore_patch_objects):
    """
    Run extra indexes on a table. This will merge the indexing results for all objects
    that a table is formed from with the current object indexes. For explanation of
    what indexes do, see the documentation for `sgr commit`.

    If the objects haven't been downloaded yet, this will download them.

    Currently reindexing objects that change other objects is unsupported and will raise
    an error. Pass `-o` to ignore these objects and only reindex supported objects.

    Image spec must be of the format ``[NAMESPACE/]REPOSITORY[:HASH_OR_TAG]``. If no tag is specified, ``HEAD`` is used.
    """
    from splitgraph.core.output import pluralise

    repository, image = image_spec
    table = image.get_table(table_name)
    click.echo("Reindexing table %s:%s/%s" % (repository.to_schema(), image.image_hash, table_name))
    reindexed = table.reindex(
        extra_indexes=index_options, raise_on_patch_objects=not ignore_patch_objects
    )
    click.echo("Reindexed %s" % pluralise("object", len(reindexed)))



from typing import Any, Dict, Optional, Tuple

from splitgraph.core.types import TableColumn, TableSchema
from splitgraph.ingestion.common import dedupe_sg_schema

try:
    from multicorn import ANY
except ImportError:
    # Don't fail the import if we're running this directly.
    ANY = object()


_socrata_types = {
    "checkbox": "boolean",
    "double": "double",
    "floating timestamp": "timestamp without time zone",
    "calendar date": "date",
    "money": "money",
    "number": "numeric",
    "text": "text",
    "url": "text",
}


def _socrata_to_pg_type(socrata_type):
    socrata_type = socrata_type.lower()
    if socrata_type in [
        "line",
        "location",
        "multiline",
        "multipoint",
        "multipolygon",
        "point",
    ]:
        return "json"
    if socrata_type in _socrata_types:
        return _socrata_types[socrata_type]
    else:
        return "text"


def socrata_to_sg_schema(metadata: Dict[str, Any]) -> Tuple[TableSchema, Dict[str, str]]:
    try:
        col_names = metadata["resource"]["columns_field_name"]
        col_types = metadata["resource"]["columns_datatype"]
    except KeyError:
        raise ValueError("Invalid Socrata metadata!")

    col_desc = metadata["resource"].get("columns_description") or [None] * len(col_names)

    # Prepend the Socrata :id column that we can order on and use as PK.
    col_names = [":id"] + col_names
    col_types = ["text"] + col_types
    col_desc = ["Socrata column ID"] + col_desc

    result = [
        TableColumn(i, n, _socrata_to_pg_type(t), False, d)
        for i, (n, t, d) in enumerate(zip(col_names, col_types, col_desc))
    ]

    # Truncate Socrata column names to 63 characters and calculate
    # a map of Splitgraph columns to Socrata columns.
    result_deduped = dedupe_sg_schema(result)

    sg_to_socrata_cols = {
        d.name: r.name for r, d in zip(result, result_deduped) if d.name != r.name
    }

    return result_deduped, sg_to_socrata_cols


def estimate_socrata_rows_width(columns, metadata, column_map=None):
    """Estimate number of rows required for a query and each row's width
    from the table metadata."""
    column_map = column_map or {}

    # We currently don't use qualifiers for this, they get passed to Socrata
    # directly. Socrata's metadata does contain some statistics for each column
    # (min, max, sum, most popular values etc) that can be used to help the planner though.

    contents = metadata["columns"][0]["cachedContents"]
    cardinality = int(contents["non_null"]) + int(contents["null"])
    column_widths = {c["fieldName"]: c.get("width", 100) for c in metadata["columns"]}

    # Socrata doesn't return the Socrata Row ID width in its metadata but we know
    # how big it usually is (row-XXXX.XXXX_XXXX).
    row_width = sum(column_widths[column_map.get(c, c)] if c != ":id" else 18 for c in columns)

    return cardinality, row_width


def _emit_col(col, column_map: Optional[Dict[str, str]] = None):
    column_map = column_map or {}
    return f"`{column_map.get(col, col)}`"


def _emit_val(val):
    if val is None:
        return "NULL"
    if not isinstance(val, (int, float)):
        val = str(val).replace("'", "''")
        return f"'{val}'"
    return str(val)


def _convert_op(op):
    if op in ["=", ">", ">=", "<", "<=", "<>", "!="]:
        return op
    if op == "~~":
        return "LIKE"
    return None


def _base_qual_to_socrata(col, op, value, column_map=None):
    soql_op = _convert_op(op)
    if not soql_op:
        return "TRUE"
    if value is None:
        if soql_op == "=":
            soql_op = "IS"
        elif soql_op in ("<>", "!="):
            soql_op = "IS NOT"
        else:
            return "TRUE"

    return f"{_emit_col(col, column_map)} {soql_op} {_emit_val(value)}"


def _qual_to_socrata(qual, column_map=None):
    if qual.is_list_operator:
        if qual.list_any_or_all == ANY:
            # Convert col op ANY([a,b,c]) into (cop op a) OR (col op b)...
            return " OR ".join(
                f"({_base_qual_to_socrata(qual.field_name, qual.operator[0], v, column_map)})"
                for v in qual.value
            )
        # Convert col op ALL(ARRAY[a,b,c...]) into (cop op a) AND (col op b)...
        return " AND ".join(
            f"({_base_qual_to_socrata(qual.field_name, qual.operator[0], v, column_map)})"
            for v in qual.value
        )
    else:
        return f"{_base_qual_to_socrata(qual.field_name, qual.operator, qual.value, column_map)}"


def quals_to_socrata(quals, column_map: Optional[Dict[str, str]] = None):
    """Convert a list of Multicorn quals to a SoQL query"""
    return " AND ".join(f"({_qual_to_socrata(q, column_map)})" for q in quals)


def cols_to_socrata(cols, column_map: Optional[Dict[str, str]] = None):
    # Don't add ":id" to the request since we're getting it anyway
    # through exclude_system_fields=false
    return ",".join(f"{_emit_col(c, column_map)}" for c in cols if c != ":id")


def sortkeys_to_socrata(sortkeys, column_map: Optional[Dict[str, str]] = None):
    if not sortkeys:
        # Always sort on ID for stable paging
        return ":id"

    clauses = []
    for key in sortkeys:
        if key.nulls_first != key.is_reversed:
            raise ValueError("Unsupported SortKey %s" % str(key))
        order = "DESC" if key.is_reversed else "ASC"
        clauses.append(f"{_emit_col(key.attname, column_map)} {order}")
    return ",".join(clauses)
